Package: mighty.metadata
Title: Manage 'CDISC' 'ADaM' Dataset Specifications in 'YAML' Format
Version: 0.1.0
Authors@R: c(
    person("Aksel", "Thomsen", , "oath@novonordisk.com", role = c("aut", "cre")),
    person("Ari", "Siggaard Knoph", , "aikp@novonordisk.com", role = "aut"),
    person("Matthew", "Phelps", , "mewp@novonordisk.com", role = "aut"),
    person("Giulia", "Pais", , "egzp@novonordisk.com", role = "aut"),
    person("Novo Nordisk A/S", role = "cph")
  )
Description: Load, validate, and manipulate Clinical Data Interchange
    Standards Consortium ('CDISC') Analysis Data Model ('ADaM') dataset
    metadata stored as 'YAML' files. Metadata files are validated against
    a JSON schema. Provides functions to inspect and modify columns,
    parameters, and row-level operations within and across 'ADaM' domains.
    Designed for use with the 'mighty' framework.
License: Apache License (>= 2)
URL: https://novonordisk-opensource.github.io/mighty.metadata/,
        https://github.com/NovoNordisk-OpenSource/mighty.metadata
BugReports: https://github.com/NovoNordisk-OpenSource/mighty.metadata/issues
Depends: R (>= 4.1)
Imports: cli, dplyr, glue, purrr, rlang, S7, S7schema (>= 0.1.1),
        stats, tibble, yaml, zephyr
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), tidyr, withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-11 16:16:55 UTC; oath
Author: Aksel Thomsen [aut, cre],
  Ari Siggaard Knoph [aut],
  Matthew Phelps [aut],
  Giulia Pais [aut],
  Novo Nordisk A/S [cph]
Maintainer: Aksel Thomsen <oath@novonordisk.com>
Repository: CRAN
Date/Publication: 2026-05-15 20:10:07 UTC
Built: R 4.5.2; ; 2026-05-15 22:50:34 UTC; unix

Package: mighty.metadata
Title: Manage 'CDISC' 'ADaM' Dataset Specifications in 'YAML' Format
Version: 0.1.0
Authors@R: c(
    person("Aksel", "Thomsen", , "oath@novonordisk.com", role = c("aut", "cre")),
    person("Ari", "Siggaard Knoph", , "aikp@novonordisk.com", role = "aut"),
    person("Matthew", "Phelps", , "mewp@novonordisk.com", role = "aut"),
    person("Giulia", "Pais", , "egzp@novonordisk.com", role = "aut"),
    person("Novo Nordisk A/S", role = "cph")
  )
Description: Load, validate, and manipulate Clinical Data Interchange
    Standards Consortium ('CDISC') Analysis Data Model ('ADaM') dataset
    metadata stored as 'YAML' files. Metadata files are validated against
    a JSON schema. Provides functions to inspect and modify columns,
    parameters, and row-level operations within and across 'ADaM' domains.
    Designed for use with the 'mighty' framework.
License: Apache License (>= 2)
URL: https://novonordisk-opensource.github.io/mighty.metadata/,
        https://github.com/NovoNordisk-OpenSource/mighty.metadata
BugReports: https://github.com/NovoNordisk-OpenSource/mighty.metadata/issues
Depends: R (>= 4.1)
Imports: cli, dplyr, glue, purrr, rlang, S7, S7schema (>= 0.1.1),
        stats, tibble, yaml, zephyr
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), tidyr, withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-11 16:16:55 UTC; oath
Author: Aksel Thomsen [aut, cre],
  Ari Siggaard Knoph [aut],
  Matthew Phelps [aut],
  Giulia Pais [aut],
  Novo Nordisk A/S [cph]
Maintainer: Aksel Thomsen <oath@novonordisk.com>
Repository: CRAN
Date/Publication: 2026-05-15 20:10:07 UTC
Built: R 4.5.2; ; 2026-05-15 22:05:04 UTC; unix

# Generated by roxygen2: do not edit by hand

export(add_column)
export(add_parameter)
export(add_row)
export(create_md_col)
export(list_columns)
export(list_parameters)
export(list_rows)
export(mighty_domain)
export(mighty_study)
export(move_column)
export(move_parameter)
export(move_row)
export(populate_core)
export(populate_sparse)
export(remove_columns)
export(remove_parameters)
export(remove_rows)
export(resolve_includes)
export(select_column)
export(select_parameter)
export(select_row)
export(update_column)
export(update_parameter)
export(update_row)
export(validate)
export(write_config)
importFrom(S7,validate)
importFrom(S7schema,write_config)
importFrom(dplyr,.data)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# mighty.metadata <a href="https://novonordisk-opensource.github.io/mighty.metadata/"><img src="man/figures/logo.png" align="right" height="139" alt="mighty.metadata website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/NovoNordisk-OpenSource/mighty.metadata/actions/workflows/check_and_co.yaml/badge.svg)](https://github.com/NovoNordisk-OpenSource/mighty.metadata/actions/workflows/check_and_co.yaml)
<!-- badges: end -->

## Overview

Clinical data scientists programming CDISC Analysis Data Model (ADaM)
datasets need specifications that are both human-readable and
machine-readable. These specifications typically live in Excel
spreadsheets that are hard to version-control, difficult to validate,
and cannot be consumed by automation tooling. mighty.metadata replaces
that workflow with YAML files.

The package loads YAML specifications, validates them against a JSON
schema, and provides a pipe-friendly R interface for inspecting and
modifying column, parameter, and row definitions. It works at two
levels: a single ADaM domain (`mighty_domain()`) or an entire study
(`mighty_study()`).

mighty.metadata is part of the mighty framework for automated ADaM
programming. It is built on [S7](https://rconsortium.github.io/S7/) and
[S7schema](https://novonordisk-opensource.github.io/S7schema/) for
robust validation and modern OOP.

## Installation

``` r
# Install from CRAN:
install.packages("mighty.metadata")

# Install the development version from GitHub:
pak::pak("NovoNordisk-OpenSource/mighty.metadata")
```

## Usage

### Load and inspect a single domain

`mighty_domain()` loads a YAML metadata file for a single ADaM dataset
and validates it against the schema on load.

``` r
library(mighty.metadata)

advs <- mighty_domain(
  file = system.file("examples", "advs.yml", package = "mighty.metadata")
)

advs
#> <mighty.metadata::mighty_domain>
#> ADVS: Vital Signs Analysis Dataset
#> Class: BASIC DATA STRUCTURE
#> Keys: USUBJID, PARAMCD, and AVISITN
```

Use `list_columns()`, `list_parameters()`, and `list_rows()` to see what
the specification contains.

``` r
list_columns(advs)
#>  [1] "STUDYID"  "USUBJID"  "SAFFL"    "TRTP"     "VISITNUM" "AVISITN" 
#>  [7] "AVISIT"   "PARAMCD"  "PARAM"    "AVAL"     "AVALC"
```

``` r
list_parameters(advs)
#> [1] "BMI"    "BMIGRP"
```

``` r
list_rows(advs)
#> [1] "BASELINE"
```

### Modify a domain

Each dimension (columns, parameters, rows) has a consistent set of
verbs: `list_*`, `add_*`, `remove_*`, `update_*`, `select_*`, and
`move_*`. These return a modified `mighty_domain` object and work
naturally with the pipe.

``` r
# Add a new column
advs |>
  add_column(id = "TRTA", label = "Actual Treatment", .pos = 3) |>
  list_columns()
#>  [1] "STUDYID"  "USUBJID"  "TRTA"     "SAFFL"    "TRTP"     "VISITNUM"
#>  [7] "AVISITN"  "AVISIT"   "PARAMCD"  "PARAM"    "AVAL"     "AVALC"
```

``` r
# Remove columns
advs |>
  remove_columns(c("VISITNUM", "AVALC")) |>
  list_columns()
#> [1] "STUDYID" "USUBJID" "SAFFL"   "TRTP"    "AVISITN" "AVISIT"  "PARAMCD"
#> [8] "PARAM"   "AVAL"
```

### Work with a full study

`mighty_study()` loads every YAML file from a directory. Each file
becomes a named `mighty_domain`. The optional `_study.yml` provides
study-level properties and `_mighty.yml` provides mighty framework
configuration.

``` r
study <- mighty_study(
  path = system.file("examples", package = "mighty.metadata")
)

study
#> <mighty.metadata::mighty_study/list/S7_object>
#> @ mighty: `external_data`
#> @ study: `study_id`
#> $ ADAE: <mighty.metadata::mighty_domain>
#> $ ADSL: <mighty.metadata::mighty_domain>
#> $ ADVS: <mighty.metadata::mighty_domain>
```

``` r
names(study)
#> [1] "ADAE" "ADSL" "ADVS"
```

``` r
str(study@study)
#> List of 1
#>  $ study_id: chr "example_study"
str(study@mighty)
#> List of 1
#>  $ external_data:List of 3
#>   ..$ :List of 2
#>   .. ..$ id  : chr "DM"
#>   .. ..$ keys: chr [1:2] "STUDYID" "USUBJID"
#>   ..$ :List of 2
#>   .. ..$ id  : chr "VS"
#>   .. ..$ keys: chr [1:2] "STUDYID" "USUBJID"
#>   ..$ :List of 2
#>   .. ..$ id  : chr "AE"
#>   .. ..$ keys: chr [1:2] "STUDYID" "USUBJID"
```

## Useful links

- `vignette("mighty-metadata")` – Getting started guide
- `vignette("adam-schema")` – Domain YAML schema reference
- `vignette("study-schema")` – Study YAML schema reference
- `vignette("mighty-schema")` – Mighty YAML schema reference
- [S7schema](https://novonordisk-opensource.github.io/S7schema/) –
  Schema validation engine used by mighty.metadata
- [Novo Nordisk Open Source R
  packages](https://novonordisk-opensource.github.io/R-packages/) –
  Overview of R packages published by Novo Nordisk.

# Helper function to mimic testthat::expect_s7_class()
# since it is not available in older R versions (e.g. 4.4.1)
if (!"expect_s7_class" %in% getNamespaceExports(ns = "testthat")) {
  expect_s7_class <- function(object, class) {
    stopifnot(inherits(x = class, what = "S7_class"))
    stopifnot(S7::S7_inherits(x = object))
    stopifnot(S7::S7_inherits(x = object, class = class))
    invisible(object)
  }
}

test_that("populate_sparse() - works with intended input", {
  study <- test_path("test_study") |>
    mighty_study() |>
    populate_core() |>
    populate_sparse() |>
    expect_no_condition() |>
    expect_s7_class(mighty_study)

  expect_equal(
    select_column(study$ADAE, "SEX")[["label"]],
    select_column(study$ADSL, "SEX")[["label"]]
  )

  study |>
    populate_sparse() |>
    expect_equal(study)
})

test_that("populate_sparse() - works on a single mighty_domain", {
  study <- test_path("test_study") |>
    mighty_study() |>
    populate_core()

  populate_sparse(study$ADAE, study = study) |>
    expect_no_condition() |>
    expect_s7_class(mighty_domain)
})

test_that("populate_sparse() - errors when predecessor does not exist", {
  study <- test_path("test_study") |>
    mighty_study() |>
    populate_core()

  i <- which_ids(study$ADSL$columns, "RACE")
  study$ADSL$columns[[i]] <- NULL

  populate_sparse(study) |>
    expect_error(
      "RACE: Predecessor `ADSL.RACE` not found in study"
    )
})

test_that("populate_sparse() - debug info for predecessors not in study (e.g. SDTM)", {
  withr::local_options(
    .new = list(mighty.metadata.verbosity_level = "debug")
  )

  study <- test_path("test_study") |>
    mighty_study() |>
    populate_core()

  populate_sparse(study) |>
    capture_messages() |>
    expect_match("Predecessor domain `[A-Z]+` not found in study")
})

test_that("populate_sparse() - errors on non-standard predecessor method", {
  study <- test_path("test_study") |>
    mighty_study() |>
    populate_core()

  i <- which_ids(study$ADAE$columns, "SEX")
  study$ADAE$columns[[i]]$method <- "invalid_method"
  study$ADAE$columns[[i]]$origin <- "Predecessor"

  populate_sparse(study) |>
    expect_error("Non standard predecessor method")
})

test_that("populate_core() adds core variables as predecessors", {
  study <- test_path("test_study") |>
    mighty_study()

  study$ADAE |>
    list_columns() |>
    intersect(c("SEX", "RACE")) |>
    expect_length(0)

  study_updated <- study |>
    populate_core() |>
    expect_no_condition() |>
    expect_s7_class(mighty_study)

  study_updated$ADAE |>
    list_columns() |>
    expect_contains(c("SEX", "RACE"))

  select_column(study_updated$ADAE, "SEX") |>
    expect_mapequal(
      list(
        id = "SEX",
        method = "ADSL.SEX",
        origin = "Predecessor"
      )
    )

  select_column(study_updated$ADAE, "RACE") |>
    expect_mapequal(
      list(
        id = "RACE",
        method = "ADSL.RACE",
        origin = "Predecessor"
      )
    )
})

test_that("populate_core() throws error when column already exists", {
  study <- test_path("test_study") |>
    mighty_study()

  i <- which_ids(x = study$ADSL$columns, id = "STUDYID")
  study$ADSL$columns[[i]][["is_core"]] <- TRUE

  populate_core(study) |>
    expect_error(
      "ADAE - Variable\\(s\\) with same name already exists: `ADSL.STUDYID`"
    )
})

test_that("populate_core() throws error with duplicate core variables", {
  study <- test_path("test_study") |>
    mighty_study()

  study$ADSL2 <- study$ADSL

  populate_core(study) |>
    expect_error(
      "Non-unique core variable\\(s\\) found: `ADSL2.SEX` and `ADSL2.RACE`"
    )
})

test_that("populate_core() throws error with non-ADSL core variables", {
  study <- test_path("test_study") |>
    mighty_study()

  study$ADSL2 <- study$ADSL
  study$ADSL2 <- remove_columns(study$ADSL2, c("SEX", "RACE"))

  i <- which(list_columns(study$ADSL2) %in% "AGE")
  study$ADSL2$columns[[i]][["is_core"]] <- TRUE

  populate_core(study) |>
    expect_error(
      "Only ADSL is allowed to have core columns. Found: `ADSL2.AGE`"
    )
})

test_that("mighty_study(populate = FALSE) does not populate", {
  study <- test_path("test_study") |>
    mighty_study(populate = FALSE) |>
    expect_no_condition() |>
    expect_s7_class(mighty_study)

  study$ADAE |>
    list_columns() |>
    intersect(c("SEX", "RACE")) |>
    expect_length(0)
})

test_that("mighty_study(populate = TRUE) populates core and sparse", {
  study <- test_path("test_study") |>
    mighty_study(populate = TRUE) |>
    expect_no_condition() |>
    expect_s7_class(mighty_study)

  study$ADAE |>
    list_columns() |>
    expect_contains(c("SEX", "RACE"))

  expect_equal(
    select_column(study$ADAE, "SEX")[["label"]],
    select_column(study$ADSL, "SEX")[["label"]]
  )
})

test_that("mighty_study()", {
  study <- test_path("test_study") |>
    mighty_study() |>
    expect_no_condition() |>
    expect_s7_class(mighty_study)

  study |>
    lapply(
      FUN = expect_s7_class,
      class = mighty_domain
    ) |>
    names() |>
    expect_equal(c("ADAE", "ADSL", "ADVS"))

  S7::prop_names(study) |>
    expect_equal(c("mighty", "study", "path"))

  expect_equal(
    object = study@mighty,
    expected = list(
      external_data = list(
        list(id = "DM", keys = c("STUDYID", "USUBJID"))
      )
    )
  )

  expect_equal(
    object = study@study,
    expected = list(
      study_id = "test_study"
    )
  )

  expect_error(
    study@study <- list(not_study_id = "missing required field"),
    "study_id"
  )

  expect_snapshot(
    x = print(study),
    transform = \(x) {
      # Robust between S7 versions
      gsub(pattern = "mighty\\.metadata::", replacement = "", x = x)
    }
  )
})

test_that("validate_path() errors when @path set to non-existent directory", {
  study <- test_path("test_study") |>
    mighty_study()

  expect_error(
    study@path <- "/nonexistent/path",
    "Directory does not exist"
  )
})

test_that("validate_path() accepts existing directory", {
  study <- test_path("test_study") |> mighty_study()
  tmp <- withr::local_tempdir()
  study@path <- tmp
  expect_equal(study@path, tmp)
})

test_that("validate_path() errors on character(0)", {
  study <- test_path("test_study") |> mighty_study()
  expect_error(study@path <- character(0), "single non-NA string")
})

test_that("validate_path() errors on NA", {
  study <- test_path("test_study") |> mighty_study()
  expect_error(study@path <- NA_character_, "single non-NA string")
})

#' Populate Predecessor Metadata
#'
#' @description
#' Populates column metadata from predecessor references. Columns with a
#' `method` in the format `domain.column` (e.g., `ADSL.USUBJID`) inherit
#' metadata from the referenced predecessor.
#'
#' @param x A [mighty_study] or [mighty_domain] object.
#' @param ... Additional arguments passed to methods.
#'
#' @return A modified [mighty_study] or [mighty_domain] with predecessor
#'   column metadata populated.
#'
#' @seealso [mighty_study], [populate_core()], [create_md_col()]
#'
#' @examples
#' study <- mighty_study(
#'   path = system.file("examples", package = "mighty.metadata")
#' )
#' study <- populate_sparse(study)
#'
#' @export
populate_sparse <- S7::new_generic(
  name = "populate_sparse",
  dispatch_args = "x"
)

#' @noRd
S7::method(populate_sparse, mighty_study) <- function(x) {
  populate_sparse_study(study = x)
}

#' @noRd
S7::method(populate_sparse, mighty_domain) <- function(x, study) {
  populate_sparse_domain(domain = x, study = study)
}

#' @noRd
populate_sparse_study <- function(study) {
  study[] <- lapply(
    X = study,
    FUN = populate_sparse_domain,
    study = study
  )

  study
}

#' @noRd
populate_sparse_domain <- function(domain, study) {
  Reduce(
    f = \(init, x) {
      update_predecessor(
        x = init,
        id = x,
        study = study
      )
    },
    x = list_predecessors(domain),
    init = domain
  )
}

#' Helper function to updated a single column list entry with its predecessor information.
#' For consistency it does not change the properties in `discard`.
#' @noRd
update_predecessor <- function(
  x,
  id,
  study,
  discard = c("id", "is_core", "method", "origin", "component", "depends")
) {
  column <- select_column(x = x, id = id)

  predecessor_pattern <- "^[a-zA-Z0-9]+\\.[a-zA-Z0-9]+$"
  if (!isTRUE(grepl(pattern = predecessor_pattern, x = column$method))) {
    cli::cli_abort(
      "{id}: Non standard predecessor method {.code {column$method}}.
      Must be in the {.code {{domain}}.{{column}}} format."
    )
  }

  pred_table <- gsub(pattern = "\\..*$", replacement = "", x = column$method)
  pred_column <- gsub(pattern = "^.*\\.", replacement = "", x = column$method)

  if (!pred_table %in% names(study)) {
    zephyr::msg_debug(
      "{column$id}: Predecessor domain {.code {pred_table}} not found in study"
    )
    return(
      update_column(
        x = x,
        id = id,
        origin = "Predecessor"
      )
    )
  }

  if (!pred_column %in% list_columns(x = study[[pred_table]])) {
    cli::cli_abort(
      "{column$id}: Predecessor {.code {column$method}} not found in study"
    )
  }

  pred <- study[[pred_table]] |>
    select_column(id = pred_column)

  updated_values <- names(pred) |>
    setdiff(y = discard) |>
    setdiff(y = names(column))

  do.call(
    what = update_column,
    args = c(
      x = list(x),
      id = list(id),
      origin = list("Predecessor"),
      pred[updated_values]
    )
  )
}

#' Resolve conditional metadata items
#'
#' Evaluates `include` fields on metadata items (domains, columns, rows, parameters)
#' and removes items where the condition evaluates to `FALSE`.
#'
#' Include conditions are R expressions with `{glue}` syntax for variable
#' substitution and evaluation. The `info` list provides the values used
#' to do this.
#'
#' @param x A [mighty_study] or [mighty_domain] object.
#' @param info `list()` Named list of values used to evaluate `include`
#'   expressions. For [mighty_study], merged into `x@study`.
#' @returns The input object with conditional items resolved.
#' @examples
#' study <- mighty_study(
#'   path = system.file("examples", package = "mighty.metadata")
#' )
#'
#' # Add a conditional column
#' study$ADVS <- update_column(
#'   study$ADVS,
#'   id = "STUDYID",
#'   include = "{study_id == 'my_study'}"
#' )
#'
#' # Condition TRUE: column kept (study_id is "my_study" in @study)
#' study |>
#'   resolve_includes() |>
#'   getElement("ADVS") |>
#'   list_columns()
#'
#' # Condition FALSE: column removed
#' study |>
#'   resolve_includes(info = list(study_id = "other")) |>
#'   getElement("ADVS") |>
#'   list_columns()
#'
#' @export
resolve_includes <- S7::new_generic(
  name = "resolve_includes",
  dispatch_args = "x",
  fun = function(x, info = list()) S7::S7_dispatch()
)

#' @noRd
S7::method(resolve_includes, mighty_study) <- function(x, info = list()) {
  includes_resolve_study(study = x, info = info)
}

#' @noRd
S7::method(resolve_includes, mighty_domain) <- function(x, info = list()) {
  includes_resolve_domain(domain = x, info = info)
}

#' @noRd
includes_resolve_study <- function(study, info = list()) {
  study@study[names(info)] <- info

  result <- lapply(
    X = study,
    FUN = resolve_includes,
    info = study@study
  )

  dropped_domains <- names(result)[purrr::map_lgl(result, is.null)]
  for (domain in dropped_domains) {
    study[[domain]] <- NULL
  }

  S7::S7_data(study) <- result |> purrr::discard(is.null)

  study
}

#' @noRd
includes_resolve_domain <- function(domain, info = list()) {
  if (
    !is.null(domain[["include"]]) && !eval_include(domain[["include"]], info)
  ) {
    return(NULL)
  }
  domain$columns <- includes_resolve_list(domain$columns, info)
  domain$rows <- includes_resolve_list(domain$rows, info)
  domain$parameters <- includes_resolve_list(domain$parameters, info)
  validate(domain)
}

#' @noRd
includes_resolve_list <- function(x, info) {
  if (is.null(x)) {
    return(NULL)
  }

  ids <- list_includes(x)

  if (length(ids) == 0L) {
    return(x)
  }

  include <- vapply(
    X = ids,
    FUN = \(id) eval_include(get_id(x, id)[["include"]], info),
    FUN.VALUE = logical(1)
  )

  x |>
    update_ids(id = ids[include], .remove = "include") |>
    remove_ids(id = ids[!include])
}

#' @noRd
eval_include <- function(include, info) {
  if (is.logical(include)) {
    return(include)
  }

  glue::glue_data(.x = info, include) |>
    as.logical() |>
    isTRUE()
}

test_that("resolve_includes() - study - simple", {
  study <- test_path("test_study") |>
    mighty_study()

  study@study$study_id |>
    expect_equal("test_study")

  study |>
    resolve_includes() |>
    expect_equal(study)

  study$ADAE <- update_column(
    study$ADAE,
    id = "TRTEMFL",
    include = "{study_id == 'test_study'}"
  )

  study |>
    resolve_includes() |>
    getElement("ADAE") |>
    list_columns() |>
    expect_contains("TRTEMFL")

  study@study$study_id <- "other_study"

  study |>
    resolve_includes() |>
    getElement("ADAE") |>
    list_columns() |>
    expect_no_match("^TRTEMFL$")

  study2 <- study |>
    resolve_includes(
      info = list(
        study_id = "test_study"
      )
    )

  study2@study$study_id |>
    expect_equal("test_study")

  study$ADAE |>
    list_columns() |>
    expect_contains("TRTEMFL")
})

test_that("resolve_includes() - study - complex", {
  study <- test_path("test_study") |>
    mighty_study()

  study$ADSL <- study$ADSL |>
    update_column(
      id = "SEX",
      include = "{study_id == 'test_study'}"
    ) |>
    update_column(
      id = "RACE",
      include = FALSE
    )

  study$ADAE <- study$ADAE |>
    update_column(
      id = "AEBODSYS",
      include = "{study_id == 'different_study'}"
    )

  study$ADVS <- study$ADVS |>
    update_row(
      id = "BASELINE",
      include = "{impute_baseline_row}"
    ) |>
    add_parameter(
      id = "BMIGRP",
      label = "BMI Group",
      columns = list(
        list(
          id = "AVALC",
          method = "Derived from BMI"
        )
      ),
      include = "{study_id == 'test_study'}"
    )

  study_resolved <- resolve_includes(
    x = study,
    info = list(
      impute_baseline_row = FALSE
    )
  )

  expect_equal(
    object = study_resolved@study,
    expected = list(
      study_id = "test_study",
      impute_baseline_row = FALSE
    )
  )

  study_resolved$ADSL |>
    list_columns() |>
    expect_contains("SEX") |>
    expect_no_match("^RACE$")

  study_resolved$ADAE |>
    list_columns() |>
    expect_no_match("^AEBODSYS$")

  study_resolved$ADVS |>
    list_rows() |>
    expect_length(0L)

  study_resolved$ADVS |>
    list_parameters() |>
    expect_contains(c("BMI", "BMIGRP"))
})

test_that("resolve_includes() - domain", {
  adsl <- test_path("test_study", "adsl.yml") |>
    mighty_domain()

  adsl |>
    resolve_includes() |>
    expect_equal(adsl)

  adsl <- adsl |>
    update_column(
      id = "SITEID",
      include = FALSE
    ) |>
    update_column(
      id = "AGE",
      include = TRUE
    )

  adsl_resolved <- adsl |>
    resolve_includes()

  adsl_resolved |>
    list_columns() |>
    expect_contains("AGE") |>
    expect_no_match("^SITEID$")

  adsl_resolved |>
    select_column(id = "AGE") |>
    names() |>
    expect_no_match("^include$")
})

test_that("resolve_includes() - drops domains if not included", {
  study <- test_path("test_study") |>
    mighty_study()

  study$ADVS[["include"]] <- FALSE

  validate(study) |>
    expect_no_condition()

  result <- study |>
    resolve_includes()

  result |>
    getElement("ADVS") |>
    expect_null()

  result |>
    getElement("ADSL") |>
    S7::S7_inherits(class = mighty_domain) |>
    expect_true()

  result |>
    getElement("ADAE") |>
    S7::S7_inherits(class = mighty_domain) |>
    expect_true()
})

test_that("resolve_includes() - drops domains if not included (expression)", {
  study <- test_path("test_study") |>
    mighty_study()

  study$ADVS[["include"]] <- "{study_id == 'test_study'}"

  validate(study) |>
    expect_no_condition()

  study |>
    resolve_includes() |>
    getElement("ADVS") |>
    S7::S7_inherits(class = mighty_domain) |>
    expect_true()

  study@study$study_id <- "other_study"

  study |>
    resolve_includes() |>
    getElement("ADVS") |>
    expect_null()
})

test_that("eval_include()", {
  info <- list(x = "a", y = 2, z = 10)

  eval_include("{x == 'a'}", info) |>
    expect_true()

  eval_include("{x == 'a' & y > 5}", info) |>
    expect_false()

  eval_include(TRUE, info) |>
    expect_true()

  eval_include(FALSE, info) |>
    expect_false()

  eval_include("{unknown_info}", info) |>
    expect_error()

  eval_include("{y + z > 10}", info) |>
    expect_true()
})

#' Update rows in your metadata
#'
#' Functions to list, select, remove, add, update, and move row operations in
#' your `mighty_domain()` objects.
#'
#' @param x `mighty_domain()` Object to manipulate.
#' @param id `character()` Id of the row(s) to remove, add, or move.
#' @param ... Additional properties to add for the row, e.g. a `label = "my row"`.
#' @param .pos `integer(1)` Position to put or move the row to. Default places the row as the last.
#' @returns `invisible(x)`
#' @examples
#' # Load example configuration
#' x <- mighty_domain(
#'   file = system.file("examples", "advs.yml", package = "mighty.metadata")
#' )
#'
#' # List all rows defined
#' list_rows(x)
#'
#' # Add a new row
#' y <- x |>
#'   add_row(id = "NEW")
#'
#' list_rows(y)
#'
#' # Remove the new row again
#' y |>
#'   remove_rows("NEW") |>
#'   list_rows()
#'
#' # Update an existing row
#' x |>
#'   update_row(id = "BASELINE", method = "Updated method") |>
#'   select_row(id = "BASELINE") |>
#'   str()
#'
#' # Select a specific row
#' select_row(x, id = "BASELINE") |>
#'   str()
#'
#' @name rows
NULL

#' @rdname rows
#' @export
list_rows <- S7::new_generic(
  name = "list_rows",
  dispatch_args = "x",
  fun = \(x) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(list_rows, mighty_domain) <- function(x) {
  row_list(x)
}

#' @rdname rows
#' @export
remove_rows <- S7::new_generic(
  name = "remove_rows",
  dispatch_args = "x",
  fun = \(x, id) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(remove_rows, mighty_domain) <- function(x, id) {
  row_remove(x, id)
}

#' @rdname rows
#' @export
add_row <- S7::new_generic(
  name = "add_row",
  dispatch_args = c("x"),
  fun = \(x, id, ..., .pos = length(x[["rows"]]) + 1L) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(add_row, mighty_domain) <- function(
  x,
  id,
  ...,
  .pos = length(x[["rows"]]) + 1L
) {
  row_add(x, id, ..., .pos = .pos)
}

#' @rdname rows
#' @export
move_row <- S7::new_generic(
  name = "move_row",
  dispatch_args = c("x"),
  fun = \(x, id, .pos = length(x[["rows"]])) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(move_row, mighty_domain) <- function(
  x,
  id,
  .pos = length(x[["rows"]])
) {
  row_move(x, id, .pos = .pos)
}

#' @rdname rows
#' @export
update_row <- S7::new_generic(
  name = "update_row",
  dispatch_args = "x",
  fun = \(x, id, ...) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(update_row, mighty_domain) <- function(x, id, ...) {
  row_update(x, id, ...)
}

#' @rdname rows
#' @export
select_row <- S7::new_generic(
  name = "select_row",
  dispatch_args = "x",
  fun = \(x, id) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(select_row, mighty_domain) <- function(x, id) {
  row_select(x, id)
}

#' @noRd
row_list <- function(x) {
  list_ids(x[["rows"]])
}

#' @noRd
row_remove <- function(x, id) {
  x[["rows"]] <- remove_ids(x[["rows"]], id)
  validate(x)
}

#' @noRd
row_add <- function(x, id, ..., .pos) {
  row <- c(id = id, rlang::list2(...))
  x[["rows"]] <- insert_in_vector(x[["rows"]], row, pos = .pos)
  validate(x)
}

#' @noRd
row_move <- function(x, id, .pos) {
  x[["rows"]] <- move_id(x[["rows"]], id, .pos)
  validate(x)
}

#' @noRd
row_update <- function(x, id, ...) {
  x[["rows"]] <- update_ids(x[["rows"]], id, ...)
  validate(x)
}

#' @noRd
row_select <- function(x, id) {
  get_id(x[["rows"]], id)
}

#' Populate Core Variables
#'
#' @description
#' Adds core variables from supplier datasets as predecessor columns to
#' datasets that use them (marked with `usecore = TRUE`).
#'
#' Note: Currently only accepts core variables from ADSL.
#'
#' @param x A [mighty_study] or [mighty_domain] object.
#' @param ... Additional arguments passed to methods.
#'
#' @return A modified [mighty_study] or [mighty_domain] with core variables
#'   added as predecessor columns.
#'
#' @seealso [mighty_study], [populate_sparse()], [create_md_col()]
#'
#' @examples
#' study <- mighty_study(
#'   path = system.file("examples", package = "mighty.metadata")
#' )
#' study <- populate_core(study)
#'
#' @export
populate_core <- S7::new_generic(
  name = "populate_core",
  dispatch_args = "x"
)

#' @noRd
S7::method(populate_core, mighty_study) <- function(x) {
  populate_core_study(study = x)
}

#' @noRd
S7::method(populate_core, mighty_domain) <- function(x, study) {
  populate_core_domain(domain = x, study = study)
}

#' @noRd
populate_core_study <- function(study) {
  study[] <- lapply(
    X = study,
    FUN = populate_core,
    study = study
  )

  study
}

#' Adds core variables from ADSL
#' and checks for duplicate column specifications.
#' Note: If we later want to allow non-ADSL core variables
#' just remove the check.
#' @noRd
populate_core_domain <- function(domain, study) {
  if (!isTRUE(domain$usecore)) {
    return(domain)
  }

  study[[which(names(study) %in% domain$id)]] <- NULL

  core_vars <- lapply(X = study, FUN = list_cores) |>
    lapply(FUN = \(x) stats::setNames(nm = x)) |>
    unlist() |>
    toupper()

  dup_core_vars <- which(duplicated(core_vars))
  if (length(dup_core_vars)) {
    cli::cli_abort(
      "Non-unique core variable(s) found: {.code {names(core_vars)[dup_core_vars]}}"
    )
  }

  non_adsl_core_vars <- which(
    !grepl(pattern = "^ADSL\\.", x = names(core_vars))
  )
  if (length(non_adsl_core_vars)) {
    cli::cli_abort(
      "Only ADSL is allowed to have core columns.
      Found: {.code {names(core_vars)[non_adsl_core_vars]}}"
    )
  }

  dup_vars <- which(core_vars %in% list_columns(domain))
  if (length(dup_vars)) {
    cli::cli_abort(
      "{domain$id} - Variable(s) with same name already exists: {.code {names(core_vars)[dup_vars]}}"
    )
  }

  predecessor <- names(core_vars)

  for (i in seq_along(core_vars)) {
    domain <- add_column(
      x = domain,
      id = core_vars[[i]],
      method = predecessor[[i]],
      origin = "Predecessor"
    )
  }

  domain
}

#' Mighty Study
#'
#' @description
#' Creates a `mighty_study` object by loading all YAML metadata files from a
#' directory. Each YAML file (except `_mighty.yml` and `_study.yml`) is parsed
#' as a [mighty_domain] object. The optional `_study.yml` file provides
#' study-level properties and the optional `_mighty.yml` file provides
#' mighty framework configuration.
#'
#' @param path `character(1)` path to a directory containing YAML metadata files.
#' @param populate `logical(1)` if `TRUE`, calls [populate_core()] then
#'   [populate_sparse()] before returning. Default is `FALSE`.
#'
#' @return A `mighty_study` S7 object extending `list`:
#' \describe{
#'   \item{List elements}{[mighty_domain] objects, named by their `id` field.
#'     Access via e.g. `study$adsl`.}
#'   \item{`@study`}{Study-level properties from `_study.yml`, or empty list
#'     if no properties file exists.}
#'   \item{`@mighty`}{Mighty framework configuration from `_mighty.yml`, or empty list
#'     if no configuration file exists.}
#'   \item{`@path`}{The source directory path as `character(1)`.}
#' }
#'
#' @details
#' The function scans the directory for files matching `*.yaml` or `*.yml`:
#' - Files named `_study.yml` or `_study.yaml` are treated as study properties
#' - Files named `_mighty.yml` or `_mighty.yaml` are treated as mighty framework config
#' - All other YAML files are loaded as [mighty_domain] objects
#' - Only one `_mighty.yml` and one `_study.yml` file is allowed per directory
#'
#' @seealso [mighty_domain], [write_config()], [populate_sparse()],
#'   [populate_core()], [create_md_col()]
#'
#' @examples
#' # Load example study
#' study <- mighty_study(
#'   path = system.file("examples", package = "mighty.metadata")
#' )
#'
#' # List tables with metadata
#' names(study)
#'
#' # Access ADVS
#' study$ADVS
#'
#' # Access study-level properties
#' study@study
#'
#' # Access mighty framework configuration
#' study@mighty
#'
#' # Load and populate in one step
#' study <- mighty_study(
#'   path = system.file("examples", package = "mighty.metadata"),
#'   populate = TRUE
#' )
#'
#' # Write study back to YAML
#' tmp <- tempdir()
#' write_config(study, path = tmp)
#'
#' @name mighty_study
NULL

#' @noRd
construct_mighty_study <- function(path, populate = FALSE) {
  mighty_schema <- system.file(
    "schema",
    "mighty.json",
    package = "mighty.metadata"
  )
  study_schema <- system.file(
    "schema",
    "study.json",
    package = "mighty.metadata"
  )

  mighty_file <- find_yml(path = path, name = "_mighty", schema = mighty_schema)
  study_file <- find_yml(path = path, name = "_study", schema = study_schema)

  entries <- list.files(
    path = path,
    pattern = "\\.(yaml|yml)$",
    full.names = TRUE
  ) |>
    setdiff(c(mighty_file, study_file))

  entries <- lapply(X = entries, FUN = mighty_domain)

  names(entries) <- vapply(
    X = entries,
    FUN = \(x) x$id,
    FUN.VALUE = character(1)
  )

  study <- S7::new_object(
    .parent = entries,
    mighty = read_yml(file = mighty_file),
    study = read_yml(file = study_file),
    path = path
  )

  if (!isTRUE(populate)) {
    return(study)
  }

  study |>
    populate_core() |>
    populate_sparse()
}

#' @noRd
validate_mighty <- function(value) {
  if (length(value) > 0) {
    schema <- system.file(
      "schema",
      "mighty.json",
      package = "mighty.metadata"
    )
    S7schema::validate_list(value, schema)
    check_unique_ids(value)
  }
  NULL
}

#' @noRd
validate_study <- function(value) {
  if (length(value) > 0) {
    schema <- system.file(
      "schema",
      "study.json",
      package = "mighty.metadata"
    )
    S7schema::validate_list(value, schema)
  }
  NULL
}

#' @noRd
validate_path <- function(value) {
  if (length(value) != 1L || is.na(value)) {
    return("@path must be a single non-NA string")
  }
  if (!dir.exists(value)) {
    return("Directory does not exist")
  }
}

#' @rdname mighty_study
#' @export
mighty_study <- S7::new_class(
  name = "mighty_study",
  parent = S7::class_list,
  properties = list(
    mighty = S7::new_property(
      class = S7::class_list,
      validator = \(value) {
        validate_mighty(value = value)
      }
    ),
    study = S7::new_property(
      class = S7::class_list,
      validator = \(value) {
        validate_study(value = value)
      }
    ),
    path = S7::new_property(
      class = S7::class_character,
      validator = \(value) {
        validate_path(value = value)
      }
    )
  ),
  constructor = construct_mighty_study
)

#' @noRd
S7::method(print, mighty_study) <- function(x, ...) {
  print_mighty_study(x)
}

#' @noRd
print_mighty_study <- function(x, ...) {
  entries <- paste0(
    "$ ",
    names(x),
    ": {.cls ",
    vapply(
      X = x,
      FUN = \(x) class(x)[[1]],
      FUN.VALUE = character(1)
    ),
    "}"
  )

  mighty <- NULL
  if (length(x@mighty)) {
    mighty <- "@ mighty: {.code {names(x@mighty)}}"
  }

  study <- NULL
  if (length(x@study)) {
    study <- "@ study: {.code {names(x@study)}}"
  }

  cli::cli_bullets(
    text = c(
      "{.cls {class(x)}}",
      mighty,
      study,
      entries
    )
  )

  invisible(x)
}

#' Update columns in your metadata
#'
#' Functions to list, select, remove, add, update, and move columns in
#' your `mighty_domain()` objects.
#'
#' @param x `mighty_domain()` Object to manipulate.
#' @param id `character()` Id of the column(s) to remove, add, or move.
#' @param ... Additional properties to add/update for the column, e.g. a `label = "my label"`.
#' @param .pos `integer(1)` Position to put or move the column to. Default places the column as the last.
#' @returns `invisible(x)`
#' @examples
#' # Load example configuration
#' x <- mighty_domain(
#'   file = system.file("examples", "advs.yml", package = "mighty.metadata")
#' )
#'
#' # List all columns defined
#' list_columns(x)
#'
#' # Remove the STUDYID and USUBJID columns
#' x |>
#'   remove_columns(c("STUDYID", "USUBJID")) |>
#'   list_columns()
#'
#' # Add a new column
#' x |>
#'   add_column(id = "NEW") |>
#'   list_columns()
#'
#' # Add new column with label and as the first column
#' y <- x |>
#'   add_column(id = "NEW", label = "My label", .pos = 1)
#'
#' list_columns(y)
#'
#' y[["columns"]][[1]] |>
#'   str()
#'
#' # Move the STUDYID column to the 3rd position
#' x |>
#'   move_column(id = "STUDYID", .pos = 3) |>
#'   list_columns()
#'
#' # Update an existing column
#' x |>
#'   update_column(id = "STUDYID", label = "Updated Label") |>
#'   select_column(id = "STUDYID") |>
#'   str()
#'
#' # Select a specific column
#' select_column(x, id = "STUDYID") |>
#'   str()
#'
#' @name columns
NULL

#' @rdname columns
#' @export
list_columns <- S7::new_generic(
  name = "list_columns",
  dispatch_args = "x",
  fun = \(x) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(list_columns, mighty_domain) <- function(x) {
  col_list(x)
}

#' @rdname columns
#' @export
remove_columns <- S7::new_generic(
  name = "remove_columns",
  dispatch_args = "x",
  fun = \(x, id) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(remove_columns, mighty_domain) <- function(x, id) {
  col_remove(x, id)
}

#' @rdname columns
#' @export
add_column <- S7::new_generic(
  name = "add_column",
  dispatch_args = c("x"),
  fun = \(x, id, ..., .pos = length(x[["columns"]]) + 1L) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(add_column, mighty_domain) <- function(
  x,
  id,
  ...,
  .pos = length(x[["columns"]]) + 1L
) {
  col_add(x, id, ..., .pos = .pos)
}

#' @rdname columns
#' @export
move_column <- S7::new_generic(
  name = "move_column",
  dispatch_args = c("x"),
  fun = \(x, id, .pos = length(x[["columns"]])) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(move_column, mighty_domain) <- function(
  x,
  id,
  .pos = length(x[["columns"]])
) {
  col_move(x, id, .pos = .pos)
}

#' @rdname columns
#' @export
select_column <- S7::new_generic(
  name = "select_column",
  dispatch_args = "x",
  fun = \(x, id) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(select_column, mighty_domain) <- function(x, id) {
  col_select(x, id)
}

#' @rdname columns
#' @export
update_column <- S7::new_generic(
  name = "update_column",
  dispatch_args = "x",
  fun = \(x, id, ...) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(update_column, mighty_domain) <- function(x, id, ...) {
  col_update(x, id, ...)
}

#' @noRd
col_list <- function(x) {
  list_ids(x[["columns"]])
}

#' @noRd
col_remove <- function(x, id) {
  x[["columns"]] <- remove_ids(x[["columns"]], id)
  validate(x)
}

#' @noRd
col_add <- function(x, id, ..., .pos) {
  column <- c(id = id, rlang::list2(...))
  x[["columns"]] <- insert_in_vector(x[["columns"]], column, pos = .pos)
  validate(x)
}

#' @noRd
col_move <- function(x, id, .pos) {
  x[["columns"]] <- move_id(x[["columns"]], id, .pos)
  validate(x)
}

#' @noRd
col_select <- function(x, id) {
  get_id(x[["columns"]], id)
}

#' @noRd
col_update <- function(x, id, ...) {
  x[["columns"]] <- update_ids(x[["columns"]], id, ...)
  validate(x)
}

#' Inserts element into existing vector.
#' Default behavior inserts `y` as the last element.
#' Position is changed with pos argument.
#' @noRd
insert_in_vector <- function(x, y, pos = length(x) + 1L) {
  stopifnot(pos > 0L && length(pos) == 1L)

  if (pos <= length(x)) {
    i <- seq(from = pos, to = length(x))
    x[i + 1L] <- x[i]
  }

  x[[pos]] <- y

  x
}

#' Helper function to retrieve entry with a certain id
#' @noRd
get_id <- function(x, id) {
  idx <- which_ids(x = x, id = id)

  if (length(idx) == 0L) {
    cli::cli_abort("Id {.code {id}} does not exist")
  }

  x[[which_ids(x = x, id = id)]]
}

#' Helper function to get all `ìd`s as a character value`
#' from all elements in a list.
#' @noRd
list_ids <- function(x) {
  if (!length(x)) {
    return(character(0))
  }

  vapply(
    X = x,
    FUN = \(x) x[["id"]],
    FUN.VALUE = character(1)
  )
}

#' Helper function to list all `core` columns
#' @noRd
list_cores <- function(x) {
  vapply(
    X = x[["columns"]],
    FUN = \(x) list(x[["id"]][isTRUE(x[["is_core"]])]),
    FUN.VALUE = list(1)
  ) |>
    unlist()
}

# Helper to check if a columns is a predecessor
#' @noRd
is_predecessor <- function(col) {
  has_predecessor_origin <- isTRUE(col[["origin"]] == "Predecessor")

  has_method_reference <- isTRUE(
    grepl(
      pattern = "^[a-zA-Z0-9]+\\.[a-zA-Z0-9]+$",
      x = col[["method"]]
    )
  )

  has_predecessor_origin || has_method_reference
}

# Helper to list all predecessor columns
#' @noRd
list_predecessors <- function(x) {
  cols <- x[["columns"]]
  is_pred <- vapply(X = cols, FUN = is_predecessor, FUN.VALUE = logical(1))
  list_ids(cols[is_pred])
}

#' Remove entries by id from a list
#' @noRd
remove_ids <- function(x, id) {
  if (!length(id)) {
    return(x)
  }

  x[which_ids(x, id)] <- NULL

  if (!length(x)) {
    return(NULL)
  }

  x
}

#' Update properties of an entry by id
#' @noRd
update_ids <- function(x, id, ..., .remove = character()) {
  if (!length(id)) {
    return(x)
  }

  updates <- rlang::list2(...)
  for (idx in which_ids(x, id)) {
    x[[idx]][names(updates)] <- updates
    x[[idx]][.remove] <- NULL
  }
  x
}

#' Move an entry to a new position
#' @noRd
move_id <- function(x, id, .pos) {
  item <- get_id(x, id)
  x <- remove_ids(x, id)
  insert_in_vector(x, item, pos = .pos)
}

#' Helper function to get the index of entries
#' with certain id value(s)
#' @noRd
which_ids <- function(x, id) {
  if (!length(x)) {
    return(integer(0))
  }

  which(list_ids(x) %in% id)
}

#' Helper to list all sub-elements with an 'include' entry
#' @noRd
list_includes <- function(x) {
  if (!length(x)) {
    return(character(0))
  }

  is_include <- vapply(
    X = x,
    FUN = \(x) "include" %in% names(x),
    FUN.VALUE = logical(1)
  )
  list_ids(x[is_include])
}

#' Helper to check that all lists have unique `id` entries
#' Note, that it checks uniqueness on the same top-level and not
#' for the entire object.
#' @noRd
check_unique_ids <- function(x) {
  if (!length(x)) {
    return(x)
  }

  flat <- unlist(x)
  flat <- flat[grepl(pattern = "^[^.]+\\.id$", x = names(flat))]

  flatter <- paste(names(flat), flat, sep = ": ")

  duplicates <- unique(flatter[duplicated(flatter)])
  names(duplicates) <- rep("x", times = length(duplicates))

  if (length(duplicates)) {
    cli::cli_abort(
      c("Duplicate `id` entries found:", duplicates)
    )
  }
}

test_that("insert_in_vector()", {
  insert_in_vector(x = 1:3, y = 4) |>
    expect_equal(1:4)

  insert_in_vector(x = 1:3, y = 4, pos = 2) |>
    expect_equal(c(1, 4, 2, 3))

  insert_in_vector(x = 1:3, y = 4, pos = 6) |>
    expect_equal(c(1:3, NA, NA, 4))

  insert_in_vector(x = 1:3, y = 4, pos = 0) |>
    expect_error()

  insert_in_vector(x = 1:3, y = 4:5, pos = 0) |>
    expect_error()
})

test_that("get_id()", {
  x <- list(
    list(id = "a", value = 1),
    list(id = "b", value = 2),
    list(id = "c", value = 3)
  )

  get_id(x, "b") |>
    expect_equal(list(id = "b", value = 2))

  get_id(x, "d") |>
    expect_error(
      regexp = "Id `d` does not exist"
    )
})

test_that("list_ids", {
  x <- list(
    list(id = "a", value = 1),
    list(id = "b", value = 2),
    list(id = "c", value = 3)
  )

  list_ids(x) |>
    expect_equal(c("a", "b", "c"))

  list_ids(list()) |>
    expect_equal(character(0))
})

test_that("which_ids", {
  x <- list(
    list(id = "a", value = 1),
    list(id = "b", value = 2),
    list(id = "c", value = 3)
  )

  which_ids(x, "a") |>
    expect_equal(1L)

  which_ids(x, c("b", "c")) |>
    expect_equal(2:3)

  which_ids(list(), "a") |>
    expect_equal(integer(0))
})

test_that("remove_ids()", {
  x <- list(list(id = "a"), list(id = "b"))

  remove_ids(x, character()) |>
    expect_equal(x)

  remove_ids(x, "a") |>
    expect_length(1L)

  remove_ids(x, c("a", "b")) |>
    expect_null()
})

test_that("update_ids()", {
  x <- list(list(id = "a", value = 1), list(id = "b", value = 2))

  update_ids(x, "a", value = 10) |>
    getElement(1) |>
    expect_equal(list(id = "a", value = 10))

  update_ids(x, character(), value = 10) |>
    expect_equal(x)
})

test_that("list_includes()", {
  x <- list(
    list(id = "a", include = "other"),
    list(id = "b", value = 1)
  )

  list_includes(x) |>
    expect_equal("a")

  list_includes(list()) |>
    expect_equal(character(0))
})

test_that("check_unique_ids()", {
  list() |>
    check_unique_ids() |>
    expect_length(0L)

  list(
    list(id = "a", x = 1),
    list(id = "b", x = 1)
  ) |>
    check_unique_ids() |>
    expect_no_condition()

  x <- list(
    id = "a",
    new_list = list(
      list(id = "a", a = 1),
      list(id = "b", b = 1),
      list(id = "a", a = 2)
    ),
    other_list = list(
      sub_list = list(
        list(id = "a"),
        list(id = "b")
      )
    )
  )

  check_unique_ids(x) |>
    expect_error("Duplicate `id` entries found")
})

#' @title Options for mighty.metadata
#' @name mighty.metadata-options
#' @description
#' `r zephyr::list_options(as = "markdown", .envir = "mighty.metadata")`
NULL

#' @title Internal parameters for reuse in functions
#' @name mighty.metadata-options-params
#' @eval zephyr::list_options(as = "params", .envir = "mighty.metadata")
#' @details
#' See [mighty.metadata-options] for more information.
#' @keywords internal
NULL

zephyr::create_option(
  name = "verbosity_level",
  default = NA_character_,
  desc = "Verbosity level for functions in mighty.metadata. See [zephyr::verbosity_level] for details." # nolint: line_length_linter
)

#' Update parameters in your metadata
#'
#' Functions to list, select, remove, add, update, and move parameters in
#' your `mighty_domain()` objects.
#'
#' @param x `mighty_domain()` Object to manipulate.
#' @param id `character()` Id of the parameter(s) to remove, add, or move.
#' @param label `character(1)` Parameter label.
#' @param columns `list()` Columns to set for the parameter.
#' @param ... Additional properties to add for the parameter, e.g. a component reference.
#' @param .pos `integer(1)` Position to put or move the parameter to. Default places the parameter as the last.
#' @returns `invisible(x)`
#' @examples
#' # Load example configuration
#' x <- mighty_domain(
#'   file = system.file("examples", "advs.yml", package = "mighty.metadata")
#' )
#'
#' # List all parameters defined
#' list_parameters(x)
#'
#' # Remove the BMIGRP parameter
#' x |>
#'   remove_parameters("BMIGRP") |>
#'   list_parameters()
#'
#' # Add a new parameter
#' x |>
#'   add_parameter(
#'     id = "NEW",
#'     label = "My new parameter",
#'     columns = list(list(id = "AVAL"))
#'   ) |>
#'   list_parameters()
#'
#' # Move the BMIGRP parameter to the 1st position
#' x |>
#'   move_parameter(id = "BMIGRP", .pos = 1) |>
#'   list_parameters()
#'
#' # Update an existing parameter
#' x |>
#'   update_parameter(id = "BMI", label = "Updated BMI Label") |>
#'   select_parameter(id = "BMI") |>
#'   str()
#'
#' # Select a specific parameter
#' select_parameter(x, id = "BMI") |>
#'   str()
#'
#' @name parameters
NULL

#' @rdname parameters
#' @export
list_parameters <- S7::new_generic(
  name = "list_parameters",
  dispatch_args = "x",
  fun = \(x) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(list_parameters, mighty_domain) <- function(x) {
  param_list(x)
}

#' @rdname parameters
#' @export
remove_parameters <- S7::new_generic(
  name = "remove_parameters",
  dispatch_args = "x",
  fun = \(x, id) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(remove_parameters, mighty_domain) <- function(x, id) {
  param_remove(x, id)
}

#' @rdname parameters
#' @export
add_parameter <- S7::new_generic(
  name = "add_parameter",
  dispatch_args = c("x"),
  fun = \(x, id, label, columns, ..., .pos = length(x[["parameters"]]) + 1L) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(add_parameter, mighty_domain) <- function(
  x,
  id,
  label,
  columns,
  ...,
  .pos = length(x[["parameters"]]) + 1L
) {
  param_add(x, id, label, columns, ..., .pos = .pos)
}

#' @rdname parameters
#' @export
move_parameter <- S7::new_generic(
  name = "move_parameter",
  dispatch_args = c("x"),
  fun = \(x, id, .pos = length(x[["parameters"]])) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(move_parameter, mighty_domain) <- function(
  x,
  id,
  .pos = length(x[["parameters"]])
) {
  param_move(x, id, .pos = .pos)
}

#' @rdname parameters
#' @export
update_parameter <- S7::new_generic(
  name = "update_parameter",
  dispatch_args = "x",
  fun = \(x, id, ...) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(update_parameter, mighty_domain) <- function(x, id, ...) {
  param_update(x, id, ...)
}

#' @rdname parameters
#' @export
select_parameter <- S7::new_generic(
  name = "select_parameter",
  dispatch_args = "x",
  fun = \(x, id) {
    S7::S7_dispatch()
  }
)

#' @noRd
S7::method(select_parameter, mighty_domain) <- function(x, id) {
  param_select(x, id)
}

#' @noRd
param_list <- function(x) {
  list_ids(x[["parameters"]])
}

#' @noRd
param_remove <- function(x, id) {
  x[["parameters"]] <- remove_ids(x[["parameters"]], id)
  validate(x)
}

#' @noRd
param_add <- function(x, id, label, columns, ..., .pos) {
  parameter <- c(
    id = id,
    label = label,
    columns = list(columns),
    rlang::list2(...)
  )
  x[["parameters"]] <- insert_in_vector(
    x[["parameters"]],
    parameter,
    pos = .pos
  )
  validate(x)
}

#' @noRd
param_move <- function(x, id, .pos) {
  x[["parameters"]] <- move_id(x[["parameters"]], id, .pos)
  validate(x)
}

#' @noRd
param_update <- function(x, id, ...) {
  x[["parameters"]] <- update_ids(x[["parameters"]], id, ...)
  validate(x)
}

#' @noRd
param_select <- function(x, id) {
  get_id(x[["parameters"]], id)
}

#' @section Write Study Metadata:
#' Use `write_config()` to serialize a `mighty_study()` object back to YAML
#' files. Each domain is written as a separate file, plus
#' `_mighty.yml` and `_study.yml` when non-empty.
#'
#' If `path` is `NULL` (default), files are written to `x@path`.
#' @name mighty_study
#' @rdname mighty_study
S7::method(write_config, mighty_study) <- function(x, path = NULL) {
  write_mighty_study(x, path)
}

#' @noRd
write_mighty_study <- function(study, path) {
  validate(study)

  if (is.null(path)) {
    path <- study@path
  }

  if (length(study@mighty)) {
    write_yml(x = study@mighty, path = path, name = "_mighty")
  }
  if (length(study@study)) {
    write_yml(x = study@study, path = path, name = "_study")
  }

  for (i in seq_along(study)) {
    write_config(
      x = study[[i]],
      path = file.path(path, basename(study[[i]]@file))
    )
  }

  invisible(study)
}

#' Mighty Domain
#'
#' @description
#' `mighty_domain()` provides a robust way of working with ADaM metadata in the `{mighty}` framework.
#'
#' A new object is initialized by supplying an existing yaml metadata file.
#' This package provides helpers to update column, parameter, and row entries.
#' See the references below for help:
#'
#' * `help("columns")`
#' * `help("parameters")`
#' * `help("rows")`
#'
#' `mighty_domain()` inherits from `S7schema::S7schema()` and the yaml file is
#' automatically validated when loaded. The helper functions above also always validates
#' the new configuration before returning.
#'
#' You can at anytime validate an object by calling `validate()` and use
#' `write_config()` to save it as a yaml file again.
#'
#' @param file `character(1)` path to a yaml file defining an ADaM dataset.
#'
#' @return A `mighty_domain` S7 object extending [S7schema::S7schema].
#'   The underlying list contains the parsed and validated YAML metadata
#'   including `id`, `label`, `class`, `keys`, `columns`, `parameters`,
#'   and `rows`.
#'
#' @examples
#' x <- mighty_domain(
#'   file = system.file("examples", "advs.yml", package = "mighty.metadata")
#' )
#'
#' # Custom print method gives a small overview
#' print(x)
#'
#' # Underlying object is a `list`
#' str(x)
#'
#' @name mighty_domain
NULL

#' @noRd
construct_mighty_domain <- function(file) {
  S7::new_object(
    .parent = S7schema::S7schema(
      file = file,
      schema = system.file("schema", "adam.json", package = "mighty.metadata")
    )
  )
}

#' @noRd
validate_mighty_domain <- function(self) {
  check_unique_ids(self)
}

#' @rdname mighty_domain
#' @export
mighty_domain <- S7::new_class(
  name = "mighty_domain",
  parent = S7schema::S7schema,
  constructor = construct_mighty_domain,
  validator = function(self) {
    validate_mighty_domain(self)
  }
)

#' @noRd
S7::method(print, mighty_domain) <- function(x, ...) {
  print_mighty_domain(x)
}

#' @noRd
print_mighty_domain <- function(x, ...) {
  cli::cli_bullets(
    text = c(
      "{.cls {class(x)[[1]]}}",
      "{x$id}: {x$label}",
      "Class: {x$class}",
      "Keys: {x$keys}"
    )
  )

  invisible(x)
}

# YAML Utilities
#
# Internal helpers for finding, reading, and writing YAML files.
#
# @param path `character(1)` Directory to search or write to.
# @param name `character(1)` File name prefix (e.g. `"_mighty"`).
# @param schema `character(1)` Path to the JSON schema file.
# @param file `character(1)` or `NULL`. Path to a YAML file.
# @param x Object to write (converted via `S7schema::to_yaml()`).

#' @noRd
list_yml <- function(path, name) {
  files <- list.files(
    path = path,
    pattern = paste0("^", name, "\\.(yaml|yml)$"),
    full.names = TRUE
  )

  if (length(files) > 1) {
    cli::cli_abort(
      "Only one {.code {name}} file allowed. Found: {.file {files}}"
    )
  }

  files
}

#' @noRd
find_yml <- function(path, name, schema) {
  files <- list_yml(path, name)

  if (length(files) == 0) {
    zephyr::msg_verbose("No {.code {name}.yml} file found")
    return(NULL)
  }

  S7schema::validate_yaml(files, schema)
}

#' @noRd
read_yml <- function(file) {
  if (is.null(file)) {
    return(list())
  }
  yaml::read_yaml(file = file)
}

#' @noRd
match_yml <- function(path, name) {
  files <- list_yml(path, name)

  if (!length(files)) {
    return(file.path(path, paste0(name, ".yml")))
  }

  files
}

#' @noRd
write_yml <- function(x, path, name) {
  cat(
    S7schema::to_yaml(x),
    file = match_yml(path, name),
    sep = ""
  )

  invisible(x)
}

test_that("find_yml() returns validated file path", {
  schema <- system.file("schema", "study.json", package = "mighty.metadata")

  result <- find_yml(
    path = test_path("test_study"),
    name = "_study",
    schema = schema
  )

  expect_type(result, "character")
  expect_length(result, 1)
  expect_true(file.exists(result))
})

test_that("find_yml() returns NULL when no file found", {
  schema <- system.file("schema", "study.json", package = "mighty.metadata")
  tmp <- withr::local_tempdir()

  find_yml(path = tmp, name = "_study", schema = schema) |>
    expect_null()
})

test_that("find_yml() errors on multiple files", {
  schema <- system.file("schema", "study.json", package = "mighty.metadata")
  tmp <- withr::local_tempdir()
  writeLines("study_id: a", file.path(tmp, "_study.yml"))
  writeLines("study_id: b", file.path(tmp, "_study.yaml"))

  expect_error(
    find_yml(path = tmp, name = "_study", schema = schema),
    "Only one `_study` file allowed"
  )
})

test_that("find_yml() errors on invalid YAML", {
  schema <- system.file("schema", "study.json", package = "mighty.metadata")
  tmp <- withr::local_tempdir()
  writeLines("not_study_id: invalid", file.path(tmp, "_study.yml"))

  expect_error(
    find_yml(path = tmp, name = "_study", schema = schema),
    "study_id"
  )
})

test_that("read_yml() returns empty list for NULL", {
  expect_equal(read_yml(file = NULL), list())
})

test_that("read_yml() returns parsed YAML for valid path", {
  file <- test_path("test_study", "_study.yml")

  result <- read_yml(file = file)

  expect_type(result, "list")
  expect_equal(result$study_id, "test_study")
})

test_that("match_yml() returns default .yml path when no file exists", {
  tmp <- withr::local_tempdir()

  match_yml(path = tmp, name = "_study") |>
    expect_equal(file.path(tmp, "_study.yml"))
})

test_that("match_yml() returns matched file when one .yml exists", {
  tmp <- withr::local_tempdir()
  writeLines("study_id: a", file.path(tmp, "_study.yml"))

  match_yml(path = tmp, name = "_study") |>
    expect_equal(file.path(tmp, "_study.yml"))
})

test_that("match_yml() returns matched file when one .yaml exists", {
  tmp <- withr::local_tempdir()
  writeLines("study_id: a", file.path(tmp, "_study.yaml"))

  match_yml(path = tmp, name = "_study") |>
    expect_equal(file.path(tmp, "_study.yaml"))
})

test_that("match_yml() errors on multiple files", {
  tmp <- withr::local_tempdir()
  writeLines("study_id: a", file.path(tmp, "_study.yml"))
  writeLines("study_id: b", file.path(tmp, "_study.yaml"))

  expect_error(
    match_yml(path = tmp, name = "_study"),
    "Only one `_study` file allowed"
  )
})

test_that("write_yml() writes YAML content to new file in empty dir", {
  tmp <- withr::local_tempdir()

  write_yml(x = list(study_id = "test"), path = tmp, name = "_study")

  expect_true(file.exists(file.path(tmp, "_study.yml")))
})

test_that("write_yml() overwrites existing .yaml file", {
  tmp <- withr::local_tempdir()
  writeLines("study_id: old", file.path(tmp, "_study.yaml"))

  write_yml(x = list(study_id = "new"), path = tmp, name = "_study")

  result <- yaml::read_yaml(file.path(tmp, "_study.yaml"))
  expect_equal(result$study_id, "new")
})

test_that("write_yml() returns invisible(x)", {
  tmp <- withr::local_tempdir()
  x <- list(study_id = "test")

  result <- write_yml(x = x, path = tmp, name = "_study")

  expect_equal(result, x)
})

#' Create Metadata Column Table
#'
#' @description
#' Converts a [mighty_study] or [mighty_domain] object into a flat dataframe
#' of column definitions.
#'
#' @param x A [mighty_study] or [mighty_domain] object.
#'
#' @return A tibble with one row per column containing:
#' \describe{
#'   \item{table_id}{Table identifier}
#'   \item{table_label}{Table label/description}
#'   \item{order}{Column order within table}
#'   \item{id}{Column name}
#'   \item{label}{Column label}
#'   \item{origin}{Origin type (e.g., "Predecessor", "Derived")}
#'   \item{key}{Logical, whether column is a key}
#'   \item{is_core}{Logical, whether column is a core variable}
#'   \item{core}{String, whether a column is Req, Cond or Perm}
#'   \item{method}{Derivation method}
#'   \item{codelist}{Codelist reference}
#'   \item{format_type}{Data type ("C" or "N")}
#'   \item{format_length}{Maximum length}
#'   \item{format_display}{Display format}
#'   \item{comment}{Comment}
#' }
#'
#' @seealso [mighty_study], [populate_sparse()], [populate_core()]
#'
#' @examples
#' study <- mighty_study(
#'   path = system.file("examples", package = "mighty.metadata")
#' )
#' mdcol <- create_md_col(study)
#'
#' @export
create_md_col <- S7::new_generic(
  name = "create_md_col",
  dispatch_args = "x",
  fun = function(x) S7::S7_dispatch()
)

#' @noRd
S7::method(create_md_col, mighty_study) <- function(x) {
  create_md_col_study(study = x)
}

#' @noRd
S7::method(create_md_col, mighty_domain) <- function(x) {
  create_md_col_domain(domain = x)
}

#' @noRd
create_md_col_study <- function(study) {
  study |>
    lapply(create_md_col) |>
    purrr::list_rbind()
}

#' @noRd
create_md_col_domain <- function(domain) {
  mdcol_cols <- names(mdcol_template)

  mdcol <- domain$columns |>
    lapply(FUN = \(x) {
      x <- purrr::list_flatten(x)
      do.call(
        what = tibble::tibble,
        args = x[names(x) %in% mdcol_cols]
      )
    }) |>
    purrr::list_rbind()

  mdcol[["table_id"]] <- domain[["id"]]
  mdcol[["table_label"]] <- domain[["label"]]
  mdcol[["key"]] <- mdcol[["id"]] %in% domain[["keys"]]
  mdcol[["order"]] <- seq_len(nrow(mdcol))

  purrr::list_rbind(list(mdcol_template, mdcol))
}

#' @noRd
mdcol_template <- tibble::tibble(
  table_id = character(),
  table_label = character(),
  order = integer(),
  id = character(),
  label = character(),
  origin = character(),
  key = logical(),
  is_core = logical(),
  core = character(),
  method = character(),
  codelist = character(),
  format_type = character(),
  format_length = integer(),
  format_display = character(),
  comment = character()
)

test_that("mighty_domain works", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  ) |>
    expect_no_condition()

  print(x) |>
    expect_snapshot(
      transform = \(x) {
        # Inconsistency between S7 versions
        sub(
          pattern = "<mighty.metadata::mighty_domain>",
          replacement = "<mighty_domain>",
          x = x
        )
      }
    )
})

test_that("list_parameters()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  list_parameters(x) |>
    expect_type("character") |>
    expect_contains("BMI")
})

test_that("remove_parameters()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  x |>
    remove_parameters(id = "BMI") |>
    list_parameters() |>
    expect_no_match("^BMI$")

  x |>
    remove_parameters(id = c("BMI", "BMIGRP")) |>
    list_parameters() |>
    expect_length(0)
})

test_that("add_parameter()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  x |>
    add_parameter(
      id = "NEW",
      label = "My new parameter",
      columns = list(
        list(
          id = "AVAL"
        )
      )
    ) |>
    list_parameters() |>
    expect_contains("NEW")

  parameters <- x |>
    add_parameter(
      id = "NEW",
      label = "My new parameter",
      columns = list(
        list(
          id = "AVAL"
        )
      ),
      .pos = 2
    ) |>
    list_parameters()

  expect_equal(parameters[[2]], "NEW")
})

test_that("move_parameter()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  ) |>
    add_parameter(
      id = "NEW",
      label = "My new parameter",
      columns = list(
        list(
          id = "AVAL"
        )
      )
    ) |>
    move_parameter(id = "NEW", .pos = 1)

  list_parameters(x)[[1]] |>
    expect_equal("NEW")

  select_parameter(x, "NEW")[["label"]] |>
    expect_equal("My new parameter")
})

test_that("update_parameter() updates existing properties", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  y <- x |>
    update_parameter(id = "BMI", label = "Updated BMI Label")

  select_parameter(y, "BMI")[["label"]] |>
    expect_equal("Updated BMI Label")
})

test_that("update_parameter() updates columns property", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  new_columns <- list(list(id = "AVAL", label = "New AVAL"))

  y <- x |>
    update_parameter(id = "BMI", columns = new_columns)

  select_parameter(y, "BMI")[["columns"]] |>
    expect_equal(new_columns)
})

test_that("select_parameter()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  parameter <- select_parameter(x, id = "BMI")
  expect_type(parameter, "list")
  expect_equal(parameter[["id"]], "BMI")
})

test_that("write_config() writes study to directory", {
  study <- test_path("test_study") |>
    mighty_study()

  tmpdir <- withr::local_tempdir()

  write_config(x = study, path = tmpdir) |>
    expect_no_error()

  list.files(tmpdir) |>
    expect_snapshot()
})

test_that("write_config() uses study@path when path is NULL", {
  tmpdir <- withr::local_tempdir()

  file.copy(list.files(test_path("test_study"), full.names = TRUE), tmpdir)

  study <- mighty_study(path = tmpdir)

  write_config(x = study) |>
    expect_no_error()

  expected <- c("_mighty.yml", "_study.yml", "adae.yml", "adsl.yml", "advs.yml")
  expect_true(all(file.exists(file.path(tmpdir, expected))))
})

test_that("write_config() roundtrip is consistent", {
  study <- test_path("test_study") |> mighty_study()
  tmpdir <- withr::local_tempdir()

  write_config(x = study, path = tmpdir)
  roundtrip <- mighty_study(path = tmpdir)

  expect_equal(roundtrip@study, study@study)
  expect_equal(roundtrip@mighty, study@mighty)
  expect_equal(names(roundtrip), names(study))
})

test_that("write_config() skips empty _mighty and _study files", {
  study <- test_path("test_study") |> mighty_study()
  tmpdir <- withr::local_tempdir()

  study@mighty <- list()
  study@study <- list()

  write_config(x = study, path = tmpdir)

  files <- list.files(tmpdir)
  expect_false("_mighty.yml" %in% files)
  expect_false("_study.yml" %in% files)
})

test_that("list_rows()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  list_rows(x) |>
    expect_type("character") |>
    expect_contains("BASELINE")
})

test_that("remove_rows()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  ) |>
    add_row(id = "NEW")

  x |>
    list_rows() |>
    expect_contains("NEW")

  x |>
    remove_rows(id = "NEW") |>
    list_rows() |>
    expect_no_match("NEW")
})

test_that("add_row()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  x |>
    add_row(
      id = "NEW",
      method = "My new row"
    ) |>
    list_rows() |>
    expect_contains("NEW")

  rows <- x |>
    add_row(
      id = "NEW",
      .pos = 1
    ) |>
    list_rows()

  expect_equal(rows[[1]], "NEW")
})

test_that("move_row()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  ) |>
    add_row(
      id = "NEW",
      method = "My new row"
    ) |>
    move_row(id = "NEW", .pos = 1)

  list_rows(x)[[1]] |>
    expect_equal("NEW")

  select_row(x, "NEW")[["method"]] |>
    expect_equal("My new row")
})

test_that("update_row() updates existing properties", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  y <- x |>
    update_row(id = "BASELINE", method = "Updated method")

  select_row(y, "BASELINE")[["method"]] |>
    expect_equal("Updated method")
})

test_that("update_row() adds new properties", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  y <- x |>
    update_row(id = "BASELINE", depends = "USUBJID")

  select_row(y, "BASELINE")[["depends"]] |>
    expect_equal("USUBJID")
})

test_that("select_row()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  row <- select_row(x, id = "BASELINE")
  expect_type(row, "list")
  expect_equal(row[["id"]], "BASELINE")
})

test_that("create_md_col()", {
  study <- test_path("test_study") |>
    mighty_study() |>
    populate_core() |>
    populate_sparse()

  mdcol <- study |>
    create_md_col() |>
    expect_no_condition() |>
    expect_s3_class("tbl_df")

  expect_equal(
    object = names(mdcol),
    expected = c(
      "table_id",
      "table_label",
      "order",
      "id",
      "label",
      "origin",
      "key",
      "is_core",
      "core",
      "method",
      "codelist",
      "format_type",
      "format_length",
      "format_display",
      "comment"
    )
  )

  expect_snapshot_value(mdcol, "json2")
})

test_that("list_columns()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  list_columns(x) |>
    expect_type("character") |>
    expect_contains("USUBJID")
})

test_that("remove_columns()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  x |>
    remove_columns(id = "USUBJID") |>
    list_columns() |>
    expect_no_match("USUBJID")

  x |>
    remove_columns(id = c("STUDYID", "USUBJID")) |>
    list_columns() |>
    expect_no_match("STUDYID|USUBJID")
})

test_that("add_column()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  x |>
    add_column(id = "NEW") |>
    list_columns() |>
    expect_contains("NEW")

  columns <- x |>
    add_column(id = "NEW", .pos = 3) |>
    list_columns()

  expect_equal(columns[[3]], "NEW")

  y <- x |>
    add_column(id = "NEW", label = "My new column") |>
    expect_no_condition()

  select_column(y, "NEW") |>
    expect_equal(
      list(id = "NEW", label = "My new column")
    )
})

test_that("move_column()", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  ) |>
    add_column(id = "NEW", label = "test") |>
    move_column(id = "NEW", .pos = 7)

  list_columns(x)[[7]] |>
    expect_equal("NEW")

  select_column(x, "NEW")[["label"]] |>
    expect_equal("test")
})

test_that("update_column() updates existing properties", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  original_label <- x[["columns"]][[1]][["label"]]

  y <- x |>
    update_column(id = "STUDYID", label = "Updated Label")

  select_column(y, "STUDYID")[["label"]] |>
    expect_equal("Updated Label")

  list_columns(y)[[1]] |>
    expect_equal("STUDYID")
})

test_that("update_column() adds new properties", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  y <- x |>
    update_column(id = "STUDYID", method = "New method value")

  select_column(y, "STUDYID")[["method"]] |>
    expect_equal("New method value")
})

test_that("update_column() preserves position", {
  x <- mighty_domain(
    file = system.file("examples", "advs.yml", package = "mighty.metadata")
  )

  original_columns <- list_columns(x)

  y <- x |>
    update_column(id = "USUBJID", label = "Updated USUBJID")

  list_columns(y) |>
    expect_equal(original_columns)
})

---
title: "Study specifications"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Study specifications}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

`_study.yml` provides study-level properties for a study directory --- the
study identifier and an optional description. The file structure is defined by a
[JSON-schema](https://json-schema.org) accessible with:

```r
system.file("schema", "study.json", package = "mighty.metadata")
```

Below you can find a detailed description of the structure.

```{r echo=FALSE}
S7schema::document_schema(
  x = system.file("schema", "study.json", package = "mighty.metadata"),
  header_start_level = 1
)
```

---
title: "ADaM specifications"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{ADaM specifications}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

`mighty.metadata` uses YAML files to specify Analysis Data Model (ADaM) datasets.
Each YAML file follows a fixed structure with defines properties and entries.

This structure is defined in a [JSON-schema](https://json-schema.org) that you can access with:

```r
system.file("schema", "adam.json", package = "mighty.metadata")
```

Below you can find a detailed description of the structure.

```{r echo=FALSE}
S7schema::document_schema(
  x = system.file("schema", "adam.json", package = "mighty.metadata"),
  header_start_level = 1
)
```

---
title: "Getting Started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, include = FALSE}
library(mighty.metadata)
```

This vignette walks you through the key workflows for defining and working with
ADaM specifications: loading and editing a single domain, assembling a
multi-domain study, propagating metadata across domains, applying conditional
includes, and producing flat output for downstream tooling.

## The YAML Specification Format

Each ADaM domain is defined in a YAML file. The simplest case is a
subject-level dataset like ADSL, which only has columns:

```yaml
id: ADSL
label: Subject-Level Analysis Dataset
class: SUBJECT LEVEL ANALYSIS DATASET
structure: One record per subject
keys: USUBJID

population:
  base:
    - domain: DM
      depends: [USUBJID]
      filter: USUBJID != ""

columns:
  - id: STUDYID
    label: Study Identifier
    method: DM.STUDYID
    core: Req
  - id: USUBJID
    label: Unique Subject Identifier
    method: DM.USUBJID
    core: Req
  # ... more columns ...
```

Top-level keys define the domain identity (`id`, `label`, `class`, `structure`,
`keys`). The `population` block declares which source domains supply raw data
and what row-level filters apply. The mighty code generator reads this to build
the initial population step; mighty.metadata stores it but does not execute it.

Each column has an `id`, `label`, and `core` conformance level: `Req`
(Required), `Cond` (Conditionally Required), or `Perm` (Permissible). The
`method` field describes how the column is derived. A `DOMAIN.COLUMN` pattern
(e.g., `DM.STUDYID`) means the column is a predecessor --- its metadata can be
inherited from the referenced source via `populate_sparse()`.

BDS (Basic Data Structure) domains like ADVS add `parameters` and `rows`:

```yaml
id: ADVS
label: Vital Signs Analysis Dataset
keys: [USUBJID, PARAMCD, AVISITN]

parameters:
  - id: BMI
    label: Body Mass Index (kg/m^2)
    columns:
      - id: AVAL
        method: Derived from height and weight

rows:
  - id: BASELINE
    method: Add baseline visit as a new row
```

See `vignette("adam-schema")` for the full schema reference.

## Working with a Single Domain

The package provides a consistent set of verbs for columns, parameters, and
rows:

| Action | Columns              | Parameters              | Rows              |
|--------|----------------------|-------------------------|--------------------|
| List   | `list_columns()`     | `list_parameters()`     | `list_rows()`      |
| Select | `select_column()`    | `select_parameter()`    | `select_row()`     |
| Add    | `add_column()`       | `add_parameter()`       | `add_row()`        |
| Update | `update_column()`    | `update_parameter()`    | `update_row()`     |
| Move   | `move_column()`      | `move_parameter()`      | `move_row()`       |
| Remove | `remove_columns()`   | `remove_parameters()`   | `remove_rows()`    |

The `remove_*` functions accept a character vector to remove multiple items at
once.

### Loading and Inspecting

Load a domain specification from a YAML file with `mighty_domain()`. The file is
validated against the ADaM JSON schema on load.

```{r}
path <- system.file("examples", "advs.yml", package = "mighty.metadata")
advs <- mighty_domain(path)
advs
```

Use `list_*()` functions to see what the specification contains:

```{r}
list_columns(advs)
list_parameters(advs)
list_rows(advs)
```

Drill into a specific column with `select_column()`:

```{r}
select_column(advs, id = "AVAL") |>
  str()
```

### Modifying Columns

Every modification automatically re-validates the domain against the schema.
All column functions return the modified domain, so they compose naturally into
a pipe chain. Here we add an actual treatment column sourced from ADSL, update
a label, and drop an unused column:

```{r}
advs <- mighty_domain(path) |>
  add_column(
    id = "TRTA",
    label = "Actual Treatment",
    method = "ADSL.TRT01A",
    .pos = 5
  ) |>
  update_column(id = "AVAL", label = "Analysis Value (Numeric)") |>
  remove_columns(id = "AVALC")

list_columns(advs)
```

#### Schema Validation

Validation runs on every modification and on initial load. You can also call
`validate()` explicitly at any time:

```{r}
validate(advs)
```

If a modification violates the schema, you get an immediate error. For example,
adding a column with a duplicate ID fails:

```{r, error = TRUE}
advs |> add_column(id = "AVAL", label = "Duplicate")
```

### Modifying Parameters

Parameters use the same verbs. The key difference is the `columns` argument in
`add_parameter()`, which accepts a nested list of column overrides specific to
that parameter:

```{r}
select_parameter(advs, id = "BMI") |>
  str()
```

```{r}
advs <- advs |>
  add_parameter(
    id = "WSTCIR",
    label = "Waist Circumference (cm)",
    columns = list(
      list(id = "AVAL", method = "VS.VSSTRESN")
    )
  )

list_parameters(advs)
```

Update and remove work as expected:

```{r}
advs <- advs |>
  update_parameter(id = "WSTCIR", label = "Waist Circumference") |>
  remove_parameters(id = "BMIGRP")

list_parameters(advs)
```

### Rows

Rows follow the same pattern. Inspect a row with `select_row()`:

```{r}
select_row(advs, id = "BASELINE") |>
  str()
```

### Saving Changes

Write the modified domain back to YAML with `write_config()`:

```{r}
out <- tempfile(fileext = ".yml")
write_config(advs, path = out)
```

The written file can be loaded back with `mighty_domain()`.

## Working with a Study

### Loading a Study

Load all domain specifications from a directory with `mighty_study()`. The
directory can contain `_study.yml` (study-level properties) and `_mighty.yml`
(mighty framework configuration).

```{r}
study_path <- system.file("examples", package = "mighty.metadata")
study <- mighty_study(study_path)
study
```

Access individual domains with `$`. Study-level properties from `_study.yml`
are stored in `@study` and mighty framework configuration from `_mighty.yml` is stored
in `@mighty`. The `@` operator accesses properties of S7 objects:

```{r}
names(study)
str(study@study)
str(study@mighty)
```

The `_study.yml` file provides the `study_id`. The `_mighty.yml` file provides
`external_data` definitions (source domains and their keys).

### Populating Core Variables

A "core variable" is an ADSL column that should appear in every consumer domain
(ADVS, ADAE, etc.) as a predecessor column --- for example, SEX and RACE for
subgroup analyses.

Note: `core` (string: Req/Cond/Perm) records the ADaM conformance level and is
unrelated to "core variable" propagation. The two fields that control
propagation are:

- `is_core` (Boolean, column-level in ADSL) --- marks a column for propagation
  to consumer domains.
- `usecore` (Boolean, domain-level) --- signals that a domain should receive the
  propagated columns. This is a top-level YAML property on the consumer domain,
  so it is set via list assignment rather than `update_column()`.

`populate_core()` reads these flags and adds the marked ADSL columns to each
consumer domain as predecessor columns.

The bundled examples do not include these fields, so we add them here to
demonstrate the workflow:

```{r}
study$ADSL <- study$ADSL |>
  update_column(id = "SEX", is_core = TRUE) |>
  update_column(id = "RACE", is_core = TRUE)

study$ADAE[["usecore"]] <- TRUE

study <- study |>
  populate_core()

list_columns(study$ADAE)
```

SEX and RACE now appear in ADAE as predecessor columns sourced from ADSL.

### Populating Predecessor Metadata

Columns that reference another domain (e.g., `method: ADSL.SAFFL`) can inherit
metadata from the referenced column. `populate_sparse()` performs this lookup
across the study, filling in only missing properties.

```{r}
# Before: SAFFL in ADVS references ADSL.SAFFL
select_column(study$ADVS, id = "SAFFL") |>
  str()
```

```{r}
study <- study |> populate_sparse()

# After: origin inherited from ADSL
select_column(study$ADVS, id = "SAFFL") |>
  str()
```

### The Complete Study Pipeline

Here is the full pipeline in one block:

```{r}
study <- mighty_study(study_path)

# Mark ADSL core variables
study$ADSL <- study$ADSL |>
  update_column(id = "SEX", is_core = TRUE) |>
  update_column(id = "RACE", is_core = TRUE)

# Mark consumer domains
study$ADAE[["usecore"]] <- TRUE

# Run the pipeline
study <- study |>
  populate_core() |>
  populate_sparse()
```

If your YAML files already include the `is_core` and `usecore` flags, the
entire pipeline collapses to a single call. Passing `populate = TRUE` is
equivalent to calling `populate_core()` then `populate_sparse()` after loading:

```{r}
study <- mighty_study(study_path, populate = TRUE)
```

### Saving a Study

Write all domain files, `_study.yml`, and `_mighty.yml` back to disk
with `write_config()`:

```{r}
out <- withr::local_tempdir()
write_config(study, path = out)
list.files(out)
```

Omit `path` to write back to the original directory (`study@path`).

## Conditional Metadata

Pooled specifications can serve multiple studies by using conditional `include`
fields. Conditions are wrapped in `{braces}` and evaluated as R expressions
(via `glue::glue_data()`) against the study's `@study` values.

```{r}
study <- mighty_study(study_path)

study$ADVS <- study$ADVS |>
  update_column(
    id = "STUDYID",
    include = "{study_id == 'example_study'}"
  )
```

When `study_id` in `@study` matches, the condition is `TRUE` and the column is
kept:

```{r}
resolved <- resolve_includes(study)
list_columns(resolved$ADVS)
```

Override with a different value and the column is removed:

```{r}
resolved <- resolve_includes(study, info = list(study_id = "other"))
list_columns(resolved$ADVS)
```

`include` works on parameters and rows too, not just columns.

## Creating a Flat Column Table

`create_md_col()` flattens the study's column specifications into a single
tibble. This is the format consumed by downstream mighty tools.

```{r}
create_md_col(study)
```

## Next Steps

This vignette covered the core workflows: `mighty_domain()` for single datasets,
`mighty_study()` for collections, `populate_core()` and `populate_sparse()` for
metadata propagation, `write_config()` for saving changes, `resolve_includes()`
for conditional specifications, and `create_md_col()` for flat output.

To learn more:

- `vignette("adam-schema")` documents the domain YAML schema reference
- `vignette("study-schema")` documents the study YAML schema reference
- `vignette("mighty-schema")` documents the mighty YAML schema reference
- The [package reference](https://novonordisk-opensource.github.io/mighty.metadata/reference/index.html)
  lists all available functions

# create_md_col()

    {
      "type": "list",
      "attributes": {
        "names": {
          "type": "character",
          "attributes": {},
          "value": ["table_id", "table_label", "order", "id", "label", "origin", "key", "is_core", "core", "method", "codelist", "format_type", "format_length", "format_display", "comment"]
        },
        "row.names": {
          "type": "integer",
          "attributes": {},
          "value": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25]
        },
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["tbl_df", "tbl", "data.frame"]
        }
      },
      "value": [
        {
          "type": "character",
          "attributes": {},
          "value": ["ADAE", "ADAE", "ADAE", "ADAE", "ADAE", "ADAE", "ADAE", "ADAE", "ADAE", "ADAE", "ADAE", "ADSL", "ADSL", "ADSL", "ADSL", "ADSL", "ADSL", "ADSL", "ADSL", "ADVS", "ADVS", "ADVS", "ADVS", "ADVS", "ADVS"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["Adverse Events Analysis Dataset", "Adverse Events Analysis Dataset", "Adverse Events Analysis Dataset", "Adverse Events Analysis Dataset", "Adverse Events Analysis Dataset", "Adverse Events Analysis Dataset", "Adverse Events Analysis Dataset", "Adverse Events Analysis Dataset", "Adverse Events Analysis Dataset", "Adverse Events Analysis Dataset", "Adverse Events Analysis Dataset", "Subject-Level Analysis Dataset", "Subject-Level Analysis Dataset", "Subject-Level Analysis Dataset", "Subject-Level Analysis Dataset", "Subject-Level Analysis Dataset", "Subject-Level Analysis Dataset", "Subject-Level Analysis Dataset", "Subject-Level Analysis Dataset", "Vital Signs Analysis Dataset", "Vital Signs Analysis Dataset", "Vital Signs Analysis Dataset", "Vital Signs Analysis Dataset", "Vital Signs Analysis Dataset", "Vital Signs Analysis Dataset"]
        },
        {
          "type": "integer",
          "attributes": {},
          "value": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 1, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, 4, 5, 6]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["STUDYID", "USUBJID", "AESEQ", "AETERM", "AEDECOD", "AEBODSYS", "ASTDT", "AENDT", "TRTEMFL", "SEX", "RACE", "STUDYID", "USUBJID", "SUBJID", "SITEID", "AGE", "AGEU", "SEX", "RACE", "STUDYID", "USUBJID", "PARAMCD", "PARAM", "AVAL", "AVISITN"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["Study Identifier", "Unique Subject Identifier", "Sequence Number", "Reported Term for the Adverse Event", "Dictionary-Derived Term", "Body System or Organ Class", "Analysis Start Date", "Analysis End Date", "Treatment Emergent Flag", "Sex", "Race", "Study Identifier", "Unique Subject Identifier", "Subject Identifier for the Study", "Study Site Identifier", "Age", "Age Units", "Sex", "Race", "Study Identifier", "Unique Subject Identifier", "Parameter Code", "Parameter", "Analysis Value", "Analysis Visit (N)"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", null, null, null, "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor", "Predecessor"]
        },
        {
          "type": "logical",
          "attributes": {},
          "value": [false, true, true, false, false, false, false, false, false, false, false, false, true, false, false, false, false, false, false, false, true, true, false, false, true]
        },
        {
          "type": "logical",
          "attributes": {},
          "value": [null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, true, true, null, null, null, null, null, null]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["Req", "Req", "Cond", "Req", "Cond", "Cond", "Cond", "Cond", "Cond", "Req", "Req", "Req", "Req", "Req", "Req", "Req", "Req", "Req", "Req", "Req", "Req", "Req", "Req", "Req", "Req"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": ["AE.STUDYID", "AE.USUBJID", "AE.AESEQ", "AE.AETERM", "AE.AEDECOD", "AE.AEBODSYS", "Derived from AE.AESTDTC", "Derived from AE.AEENDTC", "Derived based on treatment period", "ADSL.SEX", "ADSL.RACE", "DM.STUDYID", "DM.USUBJID", "DM.SUBJID", "DM.SITEID", "DM.AGE", "DM.AGEU", "DM.SEX", "DM.RACE", "VS.STUDYID", "VS.USUBJID", "VS.VSTESTCD", "VS.VSTEST", "VS.VSSTRESN", "VS.VISITNUM"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": [null, null, null, null, null, "FANCY_CDISC_CODELIST", null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null]
        },
        {
          "type": "character",
          "attributes": {},
          "value": [null, null, null, null, null, null, "date", "date", "text", null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null]
        },
        {
          "type": "integer",
          "attributes": {},
          "value": ["NA", "NA", "NA", "NA", "NA", "NA", 8, 8, 1, "NA", "NA", "NA", "NA", "NA", "NA", "NA", "NA", "NA", "NA", "NA", "NA", "NA", "NA", "NA", "NA"]
        },
        {
          "type": "character",
          "attributes": {},
          "value": [null, null, null, null, null, null, "date9.", "date9.", null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null]
        },
        {
          "type": "character",
          "attributes": {},
          "value": [null, null, null, null, null, null, null, null, "AEs occurring more than 30 days post-treatment are not flagged", null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null]
        }
      ]
    }

# mighty_study()

    Code
      print(study)
    Message
      <mighty_study/list/S7_object>
      @ mighty: `external_data`
      @ study: `study_id`
      $ ADAE: <mighty_domain>
      $ ADSL: <mighty_domain>
      $ ADVS: <mighty_domain>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mighty.metadata-options.R
\name{mighty.metadata-options}
\alias{mighty.metadata-options}
\title{Options for mighty.metadata}
\description{
\subsection{verbosity_level}{

Verbosity level for functions in mighty.metadata. See \link[zephyr:verbosity_level]{zephyr::verbosity_level} for details.
\itemize{
\item Default: \code{NA_character_}
\item Option: \code{mighty.metadata.verbosity_level}
\item Environment: \code{R_MIGHTY.METADATA_VERBOSITY_LEVEL}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mighty-metadata-package.R
\docType{package}
\name{mighty.metadata-package}
\alias{mighty.metadata}
\alias{mighty.metadata-package}
\title{mighty.metadata: Manage 'CDISC' 'ADaM' Dataset Specifications in 'YAML' Format}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Load, validate, and manipulate Clinical Data Interchange Standards Consortium ('CDISC') Analysis Data Model ('ADaM') dataset metadata stored as 'YAML' files. Metadata files are validated against a JSON schema. Provides functions to inspect and modify columns, parameters, and row-level operations within and across 'ADaM' domains. Designed for use with the 'mighty' framework.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://novonordisk-opensource.github.io/mighty.metadata/}
  \item \url{https://github.com/NovoNordisk-OpenSource/mighty.metadata}
  \item Report bugs at \url{https://github.com/NovoNordisk-OpenSource/mighty.metadata/issues}
}

}
\author{
\strong{Maintainer}: Aksel Thomsen \email{oath@novonordisk.com}

Authors:
\itemize{
  \item Ari Siggaard Knoph \email{aikp@novonordisk.com}
  \item Matthew Phelps \email{mewp@novonordisk.com}
  \item Giulia Pais \email{egzp@novonordisk.com}
}

Other contributors:
\itemize{
  \item Novo Nordisk A/S [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/z_resolve.R
\name{resolve_includes}
\alias{resolve_includes}
\title{Resolve conditional metadata items}
\usage{
resolve_includes(x, info = list())
}
\arguments{
\item{x}{A \link{mighty_study} or \link{mighty_domain} object.}

\item{info}{\code{list()} Named list of values used to evaluate \code{include}
expressions. For \link{mighty_study}, merged into \code{x@study}.}
}
\value{
The input object with conditional items resolved.
}
\description{
Evaluates \code{include} fields on metadata items (domains, columns, rows, parameters)
and removes items where the condition evaluates to \code{FALSE}.
}
\details{
Include conditions are R expressions with \code{{glue}} syntax for variable
substitution and evaluation. The \code{info} list provides the values used
to do this.
}
\examples{
study <- mighty_study(
  path = system.file("examples", package = "mighty.metadata")
)

# Add a conditional column
study$ADVS <- update_column(
  study$ADVS,
  id = "STUDYID",
  include = "{study_id == 'my_study'}"
)

# Condition TRUE: column kept (study_id is "my_study" in @study)
study |>
  resolve_includes() |>
  getElement("ADVS") |>
  list_columns()

# Condition FALSE: column removed
study |>
  resolve_includes(info = list(study_id = "other")) |>
  getElement("ADVS") |>
  list_columns()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/y_create_md_col.R
\name{create_md_col}
\alias{create_md_col}
\title{Create Metadata Column Table}
\usage{
create_md_col(x)
}
\arguments{
\item{x}{A \link{mighty_study} or \link{mighty_domain} object.}
}
\value{
A tibble with one row per column containing:
\describe{
\item{table_id}{Table identifier}
\item{table_label}{Table label/description}
\item{order}{Column order within table}
\item{id}{Column name}
\item{label}{Column label}
\item{origin}{Origin type (e.g., "Predecessor", "Derived")}
\item{key}{Logical, whether column is a key}
\item{is_core}{Logical, whether column is a core variable}
\item{core}{String, whether a column is Req, Cond or Perm}
\item{method}{Derivation method}
\item{codelist}{Codelist reference}
\item{format_type}{Data type ("C" or "N")}
\item{format_length}{Maximum length}
\item{format_display}{Display format}
\item{comment}{Comment}
}
}
\description{
Converts a \link{mighty_study} or \link{mighty_domain} object into a flat dataframe
of column definitions.
}
\examples{
study <- mighty_study(
  path = system.file("examples", package = "mighty.metadata")
)
mdcol <- create_md_col(study)

}
\seealso{
\link{mighty_study}, \code{\link[=populate_sparse]{populate_sparse()}}, \code{\link[=populate_core]{populate_core()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mighty_study.R, R/y_write_config.R
\name{mighty_study}
\alias{mighty_study}
\title{Mighty Study}
\usage{
mighty_study(path, populate = FALSE)
}
\arguments{
\item{path}{\code{character(1)} path to a directory containing YAML metadata files.}

\item{populate}{\code{logical(1)} if \code{TRUE}, calls \code{\link[=populate_core]{populate_core()}} then
\code{\link[=populate_sparse]{populate_sparse()}} before returning. Default is \code{FALSE}.}
}
\value{
A \code{mighty_study} S7 object extending \code{list}:
\describe{
\item{List elements}{\link{mighty_domain} objects, named by their \code{id} field.
Access via e.g. \code{study$adsl}.}
\item{\verb{@study}}{Study-level properties from \verb{_study.yml}, or empty list
if no properties file exists.}
\item{\verb{@mighty}}{Mighty framework configuration from \verb{_mighty.yml}, or empty list
if no configuration file exists.}
\item{\verb{@path}}{The source directory path as \code{character(1)}.}
}
}
\description{
Creates a \code{mighty_study} object by loading all YAML metadata files from a
directory. Each YAML file (except \verb{_mighty.yml} and \verb{_study.yml}) is parsed
as a \link{mighty_domain} object. The optional \verb{_study.yml} file provides
study-level properties and the optional \verb{_mighty.yml} file provides
mighty framework configuration.
}
\details{
The function scans the directory for files matching \verb{*.yaml} or \verb{*.yml}:
\itemize{
\item Files named \verb{_study.yml} or \verb{_study.yaml} are treated as study properties
\item Files named \verb{_mighty.yml} or \verb{_mighty.yaml} are treated as mighty framework config
\item All other YAML files are loaded as \link{mighty_domain} objects
\item Only one \verb{_mighty.yml} and one \verb{_study.yml} file is allowed per directory
}
}
\section{Write Study Metadata}{

Use \code{write_config()} to serialize a \code{mighty_study()} object back to YAML
files. Each domain is written as a separate file, plus
\verb{_mighty.yml} and \verb{_study.yml} when non-empty.

If \code{path} is \code{NULL} (default), files are written to \code{x@path}.
}

\examples{
# Load example study
study <- mighty_study(
  path = system.file("examples", package = "mighty.metadata")
)

# List tables with metadata
names(study)

# Access ADVS
study$ADVS

# Access study-level properties
study@study

# Access mighty framework configuration
study@mighty

# Load and populate in one step
study <- mighty_study(
  path = system.file("examples", package = "mighty.metadata"),
  populate = TRUE
)

# Write study back to YAML
tmp <- tempdir()
write_config(study, path = tmp)

}
\seealso{
\link{mighty_domain}, \code{\link[=write_config]{write_config()}}, \code{\link[=populate_sparse]{populate_sparse()}},
\code{\link[=populate_core]{populate_core()}}, \code{\link[=create_md_col]{create_md_col()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/y_parameters.R
\name{parameters}
\alias{parameters}
\alias{list_parameters}
\alias{remove_parameters}
\alias{add_parameter}
\alias{move_parameter}
\alias{update_parameter}
\alias{select_parameter}
\title{Update parameters in your metadata}
\usage{
list_parameters(x)

remove_parameters(x, id)

add_parameter(
  x,
  id,
  label,
  columns,
  ...,
  .pos = length(x[["parameters"]]) + 1L
)

move_parameter(x, id, .pos = length(x[["parameters"]]))

update_parameter(x, id, ...)

select_parameter(x, id)
}
\arguments{
\item{x}{\code{mighty_domain()} Object to manipulate.}

\item{id}{\code{character()} Id of the parameter(s) to remove, add, or move.}

\item{label}{\code{character(1)} Parameter label.}

\item{columns}{\code{list()} Columns to set for the parameter.}

\item{...}{Additional properties to add for the parameter, e.g. a component reference.}

\item{.pos}{\code{integer(1)} Position to put or move the parameter to. Default places the parameter as the last.}
}
\value{
\code{invisible(x)}
}
\description{
Functions to list, select, remove, add, update, and move parameters in
your \code{mighty_domain()} objects.
}
\examples{
# Load example configuration
x <- mighty_domain(
  file = system.file("examples", "advs.yml", package = "mighty.metadata")
)

# List all parameters defined
list_parameters(x)

# Remove the BMIGRP parameter
x |>
  remove_parameters("BMIGRP") |>
  list_parameters()

# Add a new parameter
x |>
  add_parameter(
    id = "NEW",
    label = "My new parameter",
    columns = list(list(id = "AVAL"))
  ) |>
  list_parameters()

# Move the BMIGRP parameter to the 1st position
x |>
  move_parameter(id = "BMIGRP", .pos = 1) |>
  list_parameters()

# Update an existing parameter
x |>
  update_parameter(id = "BMI", label = "Updated BMI Label") |>
  select_parameter(id = "BMI") |>
  str()

# Select a specific parameter
select_parameter(x, id = "BMI") |>
  str()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mighty.metadata-options.R
\name{mighty.metadata-options-params}
\alias{mighty.metadata-options-params}
\title{Internal parameters for reuse in functions}
\arguments{
\item{verbosity_level}{Verbosity level for functions in mighty.metadata. See \link[zephyr:verbosity_level]{zephyr::verbosity_level} for details.. Default: \code{NA_character_}.}
}
\description{
Internal parameters for reuse in functions
}
\details{
See \link{mighty.metadata-options} for more information.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/y_populate_core.R
\name{populate_core}
\alias{populate_core}
\title{Populate Core Variables}
\usage{
populate_core(x, ...)
}
\arguments{
\item{x}{A \link{mighty_study} or \link{mighty_domain} object.}

\item{...}{Additional arguments passed to methods.}
}
\value{
A modified \link{mighty_study} or \link{mighty_domain} with core variables
added as predecessor columns.
}
\description{
Adds core variables from supplier datasets as predecessor columns to
datasets that use them (marked with \code{usecore = TRUE}).

Note: Currently only accepts core variables from ADSL.
}
\examples{
study <- mighty_study(
  path = system.file("examples", package = "mighty.metadata")
)
study <- populate_core(study)

}
\seealso{
\link{mighty_study}, \code{\link[=populate_sparse]{populate_sparse()}}, \code{\link[=create_md_col]{create_md_col()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/y_rows.R
\name{rows}
\alias{rows}
\alias{list_rows}
\alias{remove_rows}
\alias{add_row}
\alias{move_row}
\alias{update_row}
\alias{select_row}
\title{Update rows in your metadata}
\usage{
list_rows(x)

remove_rows(x, id)

add_row(x, id, ..., .pos = length(x[["rows"]]) + 1L)

move_row(x, id, .pos = length(x[["rows"]]))

update_row(x, id, ...)

select_row(x, id)
}
\arguments{
\item{x}{\code{mighty_domain()} Object to manipulate.}

\item{id}{\code{character()} Id of the row(s) to remove, add, or move.}

\item{...}{Additional properties to add for the row, e.g. a \code{label = "my row"}.}

\item{.pos}{\code{integer(1)} Position to put or move the row to. Default places the row as the last.}
}
\value{
\code{invisible(x)}
}
\description{
Functions to list, select, remove, add, update, and move row operations in
your \code{mighty_domain()} objects.
}
\examples{
# Load example configuration
x <- mighty_domain(
  file = system.file("examples", "advs.yml", package = "mighty.metadata")
)

# List all rows defined
list_rows(x)

# Add a new row
y <- x |>
  add_row(id = "NEW")

list_rows(y)

# Remove the new row again
y |>
  remove_rows("NEW") |>
  list_rows()

# Update an existing row
x |>
  update_row(id = "BASELINE", method = "Updated method") |>
  select_row(id = "BASELINE") |>
  str()

# Select a specific row
select_row(x, id = "BASELINE") |>
  str()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/y_columns.R
\name{columns}
\alias{columns}
\alias{list_columns}
\alias{remove_columns}
\alias{add_column}
\alias{move_column}
\alias{select_column}
\alias{update_column}
\title{Update columns in your metadata}
\usage{
list_columns(x)

remove_columns(x, id)

add_column(x, id, ..., .pos = length(x[["columns"]]) + 1L)

move_column(x, id, .pos = length(x[["columns"]]))

select_column(x, id)

update_column(x, id, ...)
}
\arguments{
\item{x}{\code{mighty_domain()} Object to manipulate.}

\item{id}{\code{character()} Id of the column(s) to remove, add, or move.}

\item{...}{Additional properties to add/update for the column, e.g. a \code{label = "my label"}.}

\item{.pos}{\code{integer(1)} Position to put or move the column to. Default places the column as the last.}
}
\value{
\code{invisible(x)}
}
\description{
Functions to list, select, remove, add, update, and move columns in
your \code{mighty_domain()} objects.
}
\examples{
# Load example configuration
x <- mighty_domain(
  file = system.file("examples", "advs.yml", package = "mighty.metadata")
)

# List all columns defined
list_columns(x)

# Remove the STUDYID and USUBJID columns
x |>
  remove_columns(c("STUDYID", "USUBJID")) |>
  list_columns()

# Add a new column
x |>
  add_column(id = "NEW") |>
  list_columns()

# Add new column with label and as the first column
y <- x |>
  add_column(id = "NEW", label = "My label", .pos = 1)

list_columns(y)

y[["columns"]][[1]] |>
  str()

# Move the STUDYID column to the 3rd position
x |>
  move_column(id = "STUDYID", .pos = 3) |>
  list_columns()

# Update an existing column
x |>
  update_column(id = "STUDYID", label = "Updated Label") |>
  select_column(id = "STUDYID") |>
  str()

# Select a specific column
select_column(x, id = "STUDYID") |>
  str()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/y_populate_sparse.R
\name{populate_sparse}
\alias{populate_sparse}
\title{Populate Predecessor Metadata}
\usage{
populate_sparse(x, ...)
}
\arguments{
\item{x}{A \link{mighty_study} or \link{mighty_domain} object.}

\item{...}{Additional arguments passed to methods.}
}
\value{
A modified \link{mighty_study} or \link{mighty_domain} with predecessor
column metadata populated.
}
\description{
Populates column metadata from predecessor references. Columns with a
\code{method} in the format \code{domain.column} (e.g., \code{ADSL.USUBJID}) inherit
metadata from the referenced predecessor.
}
\examples{
study <- mighty_study(
  path = system.file("examples", package = "mighty.metadata")
)
study <- populate_sparse(study)

}
\seealso{
\link{mighty_study}, \code{\link[=populate_core]{populate_core()}}, \code{\link[=create_md_col]{create_md_col()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mighty_domain.R
\name{mighty_domain}
\alias{mighty_domain}
\title{Mighty Domain}
\usage{
mighty_domain(file)
}
\arguments{
\item{file}{\code{character(1)} path to a yaml file defining an ADaM dataset.}
}
\value{
A \code{mighty_domain} S7 object extending \link[S7schema:S7schema]{S7schema::S7schema}.
The underlying list contains the parsed and validated YAML metadata
including \code{id}, \code{label}, \code{class}, \code{keys}, \code{columns}, \code{parameters},
and \code{rows}.
}
\description{
\code{mighty_domain()} provides a robust way of working with ADaM metadata in the \code{{mighty}} framework.

A new object is initialized by supplying an existing yaml metadata file.
This package provides helpers to update column, parameter, and row entries.
See the references below for help:
\itemize{
\item \code{help("columns")}
\item \code{help("parameters")}
\item \code{help("rows")}
}

\code{mighty_domain()} inherits from \code{S7schema::S7schema()} and the yaml file is
automatically validated when loaded. The helper functions above also always validates
the new configuration before returning.

You can at anytime validate an object by calling \code{validate()} and use
\code{write_config()} to save it as a yaml file again.
}
\examples{
x <- mighty_domain(
  file = system.file("examples", "advs.yml", package = "mighty.metadata")
)

# Custom print method gives a small overview
print(x)

# Underlying object is a `list`
str(x)

}
